from __future__ import annotations

import enum
import importlib.resources
import json
import logging
import os
import shutil
import time
import urllib.parse
import urllib.request
from pathlib import Path, PosixPath
from typing import ClassVar, Literal
from urllib.parse import urljoin

import attrs
import pooch
from ruamel.yaml import YAML

from .._version import _version
from ..attrs import define, documented
from ..config import settings
from ..units import unit_registry as ureg
from ..util.misc import dirsize

logger = logging.getLogger(__file__)
_yaml = YAML()


class ResourceState(enum.Flag):
    NONE = 0
    CACHED = enum.auto()
    UNPACKED = enum.auto()
    INSTALLED = enum.auto()

    @staticmethod
    def to_string(value) -> str:
        result = [
            "c" if value & ResourceState.CACHED else "-",
            "u" if value & ResourceState.UNPACKED else "-",
            "i" if value & ResourceState.INSTALLED else "-",
        ]
        return "".join(result)


@define
class Resource:
    keyword: str
    hash: str
    type: str
    size: int

    #: Maps archive types to corresponding file extensions
    _EXTENSIONS: ClassVar[dict] = {"zip": "zip", "tar.gz": "tar.gz"}

    def filename(self) -> str:
        return f"{PosixPath(self.keyword)}.{Resource._EXTENSIONS[self.type]}"


@define
class AssetManager:
    """
    Lightweight package manager for Eradiate's data files.

    This class implements a basic package management system which can list,
    download, decompress and install *resources* that take the form of
    compressed archive files.

    Notes
    -----
    By default, the unique asset manager instance
    :data:`eradiate.asset_manager <.asset_manager>` is initialized based on
    user-specified configuration:

    * ``cache_dir`` ⇒ ``<settings["data_path"]>/cached``
    * ``unpack_dir`` ⇒ ``<settings["data_path"]>/unpacked``
    * ``install_dir`` ⇒ ``<settings["data_path"]>/installed/eradiate-v<__version__>``
    * ``base_uri`` ⇒ ``settings["data_url"]``

    Examples
    --------
    A single instance of the asset manager is available as
    :data:`eradiate.asset_manager <.asset_manager>`:

    >>> from eradiate import asset_manager
    >>> asset_manager.update()
    >>> asset_manager.install(["core", "panellus"])
    """

    cache_dir: Path = documented(
        attrs.field(converter=lambda x: Path(x).expanduser().resolve()),
        doc="Cache location on disk.",
        type="Path",
        init_type="path-like",
    )

    unpack_dir: Path = documented(
        attrs.field(converter=lambda x: Path(x).expanduser().resolve()),
        doc="Resource unpacking location on disk.",
        type="Path",
    )

    install_dir: Path = documented(
        attrs.field(converter=lambda x: Path(x).expanduser().resolve()),
        doc="Install location on disk.",
        type="Path",
        init_type="path-like",
    )

    base_uri: str = documented(
        attrs.field(converter=str),
        doc="Base URI of remote storage.",
        type="str or path",
    )

    #: Manifest file contents, as a dictionary.
    _manifest: dict | None = attrs.field(default=None, repr=False)

    #: Resource alias list, as a dictionary.
    _resource_aliases: dict = attrs.field(factory=dict, repr=False)

    #: List of installed resources, as a set.
    _installed: set = attrs.field(factory=set, repr=False)

    @property
    def manifest_path(self) -> Path:
        """
        Path to the manifest file.
        """
        return self.cache_dir / "manifest.json"

    @property
    def installed_path(self) -> Path:
        """
        Path to the list of installed resources.
        """
        return self.install_dir / "installed.json"

    def __attrs_post_init__(self):
        # Make sure that paths point to defined directories
        for path in [self.cache_dir, self.unpack_dir, self.install_dir]:
            path.mkdir(parents=True, exist_ok=True)

        # Load metadata files
        self._manifest = self._load_manifest()
        self._resource_aliases = self._load_resource_aliases()
        self._installed = self._load_installed()

    def _cache_size(self) -> int:
        return dirsize(self.cache_dir)

    def _unpack_size(self) -> int:
        return dirsize(self.unpack_dir)

    def _update_manifest(self) -> None:
        """
        Update cached manifest file from the remote resource store.
        """
        url = urljoin(self.base_uri, "manifest.json")
        print(f"Downloading manifest from {url}")
        urllib.request.urlretrieve(url, self.manifest_path)

    def _load_manifest(self) -> dict | None:
        """
        Read and parse manifest from disk.
        """
        manifest_path = self.manifest_path

        if manifest_path.exists():
            with open(manifest_path) as f:
                manifest = json.load(f)

            return {
                resource_id: Resource(**manifest["resources"][resource_id])
                for resource_id in manifest["resources"]
            }
        else:
            return None

    def _load_installed(self) -> set:
        """
        Read and parse list of installed resources from disk.
        """
        installed_path = self.installed_path

        if installed_path.exists():
            with open(installed_path) as f:
                installed = json.load(f)
            return set(installed)
        else:
            return set()

    @staticmethod
    def _load_resource_aliases() -> dict:
        """
        Read resource alias mapping from disk.
        """
        fname = Path(importlib.resources.files("eradiate.data") / "resources.yml")
        return _yaml.load(fname)["resources"]

    def _resolve_alias(self, resource_ids: str | list[str]) -> list[str]:
        """
        Resolve one or several resource aliases. Aliases are recursively
        resolved until the referenced resource is no longer an alias (and
        presumably points to a resource that can be found in the manifest).

        Parameters
        ----------
        resource_ids : str or list of str
            One or several resource IDs or aliases.

        Returns
        -------
        list of str
            A list of resolved resource IDs.
        """
        if isinstance(resource_ids, str):
            resource_ids = [resource_ids]
        result = []

        for resource_id in resource_ids:
            if resource_id not in self._resource_aliases:
                result.append(resource_id)
            else:
                result.extend(self._resolve_alias(self._resource_aliases[resource_id]))

        return result

    def _resolve_resource(self, resource_ids: str | list[str]) -> list[Resource]:
        """
        Return a resource definition given its ID.
        """
        resource_ids = self._resolve_alias(resource_ids)
        result = []

        for resource_id in resource_ids:
            result.append(self._manifest[resource_id])

        return result

    def _get_path_remote(self, resource: str | Resource) -> str:
        """
        Get URL to a given resource file on the remote storage.
        """
        if isinstance(resource, str):
            resource = self._manifest[resource]
        fname = resource.filename()
        return urljoin(self.base_uri, str(fname))

    def _get_path_cache(self, resource: str | Resource) -> Path:
        """
        Get cache path for a given resource.
        """
        if isinstance(resource, str):
            resource = self._manifest[resource]
        fname = resource.filename()
        return self.cache_dir / Path(fname)

    def _get_path_unpack(self, resource: str | Resource) -> Path:
        """
        Get unpack path for a given resource.
        """
        if isinstance(resource, str):
            resource = self._manifest[resource]
        dirname = PosixPath(resource.keyword)
        return self.unpack_dir / Path(dirname)

    def _list_files(self, resource_id: str) -> list[Path]:
        """
        List unpacked files for a given resource.
        """
        resource_dir = self._get_path_unpack(resource_id)
        result = []

        for root, dirs, files in os.walk(str(resource_dir)):
            for filename in files:
                result.append(Path(root) / filename)

        return result

    def _install_remove(
        self, resource_id: str, mode: Literal["install", "remove"]
    ) -> None:
        """
        Install or remove a single resource.
        """
        file_list = self._list_files(resource_id)
        unpack_root = self._get_path_unpack(resource_id)

        if mode == "install":
            print(f"Installing resource '{resource_id}'")
            for src in file_list:
                dst = self.install_dir / src.relative_to(unpack_root)
                dst.parent.mkdir(parents=True, exist_ok=True)
                if not dst.exists():
                    dst.symlink_to(src)
            self._installed.add(resource_id)

        elif mode == "remove":
            print(f"Removing resource '{resource_id}'")
            for src in file_list:
                dst = self.install_dir / src.relative_to(unpack_root)
                dst.parent.mkdir(parents=True, exist_ok=True)
                if dst.exists():
                    dst.unlink()
            if resource_id in self._installed:
                self._installed.remove(resource_id)

        else:
            raise ValueError(f"unknown mode '{mode}'")

        with open(self.installed_path, "w") as f:
            json.dump(list(self._installed), f)

    # --------------------------------------------------------------------------
    #                              Public interface
    # --------------------------------------------------------------------------

    def update(self, download: bool | None = None) -> None:
        """
        Update the manifest, either reading it from disk if offline or recent
        enough, or downloading it from remote storage.

        Parameters
        ----------
        download : bool, optional
            If ``True``, download the manifest from remote storage first.
            If ``False``, only read from disk.
            If ``None`` (unset), apply default policy (download if the file on
            disk is more than a day old and not in offline mode).
        """
        if download is None:
            if self.manifest_path.exists():
                manifest_age = time.time() - os.path.getmtime(self.manifest_path)
            else:
                manifest_age = float("inf")
            download = not settings.get("OFFLINE") and (manifest_age > 86400.0)

        if download:
            self._update_manifest()

        self._manifest = self._load_manifest()

    def info(self, show: bool = False):
        """
        Collect information about the asset manager.

        Parameters
        ----------
        show : bool
            If ``True``, display information to the terminal. Otherwise, return
            it as a dictionary.
        """

        cache_size = (self._cache_size() * ureg("B")).to_compact()
        unpack_size = (self._unpack_size() * ureg("B")).to_compact()

        result = {
            "remote_url": self.base_uri,
            "cache_dir": self.cache_dir,
            "cache_size": cache_size,
            "unpack_dir": self.unpack_dir,
            "unpack_size": unpack_size,
            "install_dir": self.install_dir,
        }

        if show:
            title = "Asset manager"
            print(title)
            print("-" * len(title))
            print(f"• Remote storage URL: {self.base_uri}")
            print(f"• Asset cache location [{cache_size:.3g~P}]: {self.cache_dir}")
            print(f"• Unpacked asset location [{unpack_size:.3g~P}]: {self.unpack_dir}")
            print(f"• Installation location: {self.install_dir}")
            return None

        else:
            return result

    def state(self, resource_ids: str | list[str]) -> dict[str, ResourceState]:
        """
        Return information about the state of one or several resources.

        Each resource has the following possible states:

        * cached: the archive is downloaded and found in the cache directory;
        * unpacked: the archive is extracted to the unpack location;
        * installed: the resource is installed to the installation directory.

        Parameters
        ----------
        resource_ids : str or list of str
            One or several resource IDs. Aliases are resolved.

        Returns
        -------
        dict
            A dictionary mapping resolved resource IDs to corresponding
            :class:`.ResourceState` flags.
        """
        if isinstance(resource_ids, str):
            resource_ids = [resource_ids]
        resource_ids = self._resolve_alias(resource_ids)
        result = {}

        for resource_id in resource_ids:
            state = ResourceState.NONE

            path_cache = self._get_path_cache(resource_id)
            if path_cache.exists():
                state |= ResourceState.CACHED

            path_unpack = self._get_path_unpack(resource_id)
            if path_unpack.exists():
                state |= ResourceState.UNPACKED

            if resource_id in self._installed:
                state |= ResourceState.INSTALLED

            result[resource_id] = state

        return result

    def list(
        self,
        what: Literal["resources", "aliases", "all"]
        | list[Literal["resources", "aliases"]] = "resources",
    ) -> None:
        """
        Display the list available resources in the manifest and their state.

        Parameters
        ----------
        what : {"resources", "aliases", "all"}
            One or several keywords that specify what to show.
        """
        from rich import box
        from rich.console import Console
        from rich.table import Table

        console = Console()

        if what == "all":
            what = ["resources", "aliases"]

        if "resources" in what:
            # List available resources and their state
            manifest = {} if self._manifest is None else self._manifest
            available_resources = []
            for resource_id, resource in manifest.items():
                kw = resource.keyword
                size = (resource.size * ureg("B")).to_compact()
                type = resource.type
                state = ResourceState.to_string(self.state(resource_id)[resource_id])
                available_resources.append((kw, type, size, state))

            available_resources.sort(key=lambda x: x[0])

            table_resources = Table(box=box.SIMPLE)
            table_resources.add_column("Resource ID")
            table_resources.add_column("Type")
            table_resources.add_column("Size")
            table_resources.add_column("State")

            for x in available_resources:
                table_resources.add_row(x[0], x[1], f"{x[2]:.3g~P}", x[3])
            console.print(table_resources)

        if "aliases" in what:
            # List and render resource aliases
            aliases = self._resource_aliases
            table_aliases = Table(box=box.SIMPLE)
            table_aliases.add_column("Alias")
            table_aliases.add_column("Target")

            for k, v in aliases.items():
                if not isinstance(v, (list, tuple)):
                    v = [v]

                for i, x in enumerate(v):
                    if i == 0:
                        table_aliases.add_row(k, x)
                    else:
                        table_aliases.add_row("", x)
            console.print(table_aliases)

    def download(
        self,
        resource_ids: str | list[str],
        unpack: bool = True,
        progressbar: bool = True,
    ) -> None:
        """
        Download a resource from remote storage to the cache directory.

        Parameters
        ----------
        resource_ids : str or list of str
            Resource ID(s) to download.

        unpack : bool, default: True
            Unpack downloaded archive.

        progressbar : bool, default: True
            If ``True``, display a progress bar for the download task.

        Notes
        -----
        This function uses the :func:`pooch.retrieve` function under the hood.
        """
        resources = self._resolve_resource(resource_ids)

        for resource in resources:
            fname = resource.filename()
            url = urljoin(self.base_uri, fname)

            if unpack:
                if resource.type == "zip":
                    processor = pooch.processors.Unzip(
                        extract_dir=self._get_path_unpack(resource).parent
                    )
                elif resource.type == "tar.gz":
                    processor = pooch.processors.Untar(
                        extract_dir=self._get_path_unpack(resource).parent
                    )
                else:
                    raise ValueError(
                        f"cannot unpack unhandled resource type '{resource.type}'"
                    )
            else:
                processor = None

            pooch.retrieve(
                url,
                fname=fname,
                path=self.cache_dir,
                known_hash=f"md5:{resource.hash}",
                processor=processor,
                progressbar=progressbar,
            )

    def install(self, resource_ids: str | list[str], progressbar: bool = True) -> None:
        """
        Install a resource and make it available for consumption.

        Parameters
        ----------
        resource_ids : str or list of str
            Resource ID(s) to install. The system ensures that requested
            resources are downloaded and unpacked.

        progressbar : bool, default: True
            If ``True``, display a progress bar if a download task is triggered.
        """
        self.update()

        to_download = [
            k
            for k, v in self.state(resource_ids).items()
            if not (v & ResourceState.UNPACKED)
        ]
        self.download(to_download, unpack=True, progressbar=progressbar)
        if isinstance(resource_ids, str):
            resource_ids = [resource_ids]
        resource_ids = self._resolve_alias(resource_ids)

        for resource_id in resource_ids:
            self._install_remove(resource_id, mode="install")

    def remove(self, resource_ids: str | list[str]) -> None:
        """
        Uninstall a single resource.

        Parameters
        ----------
        resource_ids : str or list of str
            Resource ID(s) to uninstall.
        """
        if isinstance(resource_ids, str):
            resource_ids = [resource_ids]
        resource_ids = self._resolve_alias(resource_ids)

        for resource_id in resource_ids:
            self._install_remove(resource_id, mode="remove")

    def clear(
        self,
        resource_ids: str | list[str] | None = None,
        what: Literal["cached", "unpacked", "installed", "all"]
        | list[Literal["cached", "unpacked", "installed"]] = "cached",
    ) -> None:
        """
        Clear the cache directory and save up disk space.

        Parameters
        ----------
        resource_ids : str or list of str, optional
            Resource(s) for which to clear data. If unset, all data will be
            wiped.

        what : {"cached", "unpacked", "installed", "all"}, optional
            One or several keywords that specify which part of the data to
            clear.
        """
        if isinstance(resource_ids, str):
            resource_ids = [resource_ids]
        if resource_ids is not None:
            resource_ids = self._resolve_alias(resource_ids)

        if what == "all":
            what = ["cached", "unpacked", "installed"]

        if "installed" in what:
            if resource_ids is not None:
                self.remove(resource_ids)
            else:
                print("Clearing all installed resources")
                for path in self.install_dir.iterdir():
                    if path.is_dir():
                        shutil.rmtree(path)
                self._installed.clear()
                with open(self.installed_path, "w") as f:
                    json.dump(list(self._installed), f)

        if "unpacked" in what:
            if resource_ids is not None:
                for resource_id in resource_ids:
                    print("Removing unpacked resource", resource_id)
                    path = self._get_path_unpack(resource_id)
                    if path.is_dir():
                        shutil.rmtree(path, ignore_errors=True)
            else:
                print("Clearing all unpacked resources")
                for fname in self.unpack_dir.iterdir():
                    if fname.is_dir():
                        shutil.rmtree(fname, ignore_errors=True)

        if "cached" in what:
            if resource_ids is not None:
                for resource_id in resource_ids:
                    print("Removing cached resource", resource_id)
                    path = self._get_path_cache(resource_id)
                    path.unlink(missing_ok=True)
            else:
                print("Clearing all cache")
                for fname in self.cache_dir.iterdir():
                    if fname.is_dir():
                        shutil.rmtree(fname, ignore_errors=True)


#: Unique asset manager instance (exposed as :data:`eradiate.asset_manager`).
asset_manager = AssetManager(
    cache_dir=Path(settings["data_path"]) / "cached",
    unpack_dir=Path(settings["data_path"]) / "unpacked",
    install_dir=Path(settings["data_path"]) / "installed" / f"eradiate-v{_version}",
    base_uri=settings["data_url"],
)



from __future__ import annotations

import typing as t
import warnings

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttrs
import xarray as xr

from ._core import Surface
from ..bsdfs import BSDF, LambertianBSDF, OpacityMaskBSDF
from ..core import SceneElement
from ..geometry import PlaneParallelGeometry, SceneGeometry, SphericalShellGeometry
from ..shapes import (
    BufferMeshShape,
    FileMeshShape,
    RectangleShape,
    Shape,
    SphereShape,
    shape_factory,
)
from ...attrs import define, documented, get_doc
from ...constants import EARTH_RADIUS
from ...units import symbol, to_quantity
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


def _middle(a, axis=None):
    a_min = a.min(axis=axis)
    a_max = a.max(axis=axis)
    return 0.5 * (a_min + a_max)


def _apply(transform: "mitsuba.ScalarTransform4f", vertices):
    """
    Apply a Mitsuba transform to a Numpy buffer. This function works also with
    scalar variants of Mitsuba.
    """
    return (
        transform.matrix.numpy()
        @ np.concatenate((vertices, np.ones((vertices.shape[0], 1))), axis=1).T
    ).T[:, :3]


def _mercator(lon, lat, planet_radius):
    """
    Convert longitude and latitude values to (x, y) coordinates using a
    Mercator projection.
    """
    # See math here: https://en.wikipedia.org/wiki/Mercator_projection
    x = planet_radius * lon
    y = planet_radius * np.log(np.tan(0.25 * np.pi + 0.5 * lat))
    return x, y


def _mercator_inverse(x, y, planet_radius):
    """
    Convert (x, y) coordinates to longitude and latitude using an inverse
    Mercator projection.
    """
    # See math here: https://en.wikipedia.org/wiki/Mercator_projection
    lon = x / planet_radius
    lat = 2.0 * np.arctan(np.exp(y / planet_radius)) - 0.5 * np.pi
    return lon, lat


def _transform_vertices_spherical_shell_lonlat(vertices, planet_radius):
    """
    Convert the (lon, lat, elevation) vertices from the initial vertex generation
    into (x, y, z) values for spherical shell geometries.

    Parameters
    ----------
    vertices : ndarray
        List of mesh vertices in (lon, lat, elevation) tuples, with lon and lat
        in radians.

    planet_radius : float
        Planet radius in kernel length units.

    Returns
    -------
    vertices : ndarray
    """
    lon = vertices[:, 0]
    lat = vertices[:, 1]
    lon_center, lat_center = _middle(vertices[:, :2], axis=0)
    elevation = vertices[:, 2]

    phi_r = lon
    theta_r = np.pi / 2.0 - lat

    x = np.sin(theta_r) * np.cos(phi_r) * (elevation + planet_radius)
    y = np.sin(theta_r) * np.sin(phi_r) * (elevation + planet_radius)
    z = np.cos(theta_r) * (elevation + planet_radius)
    vertices = np.array((x, y, z)).transpose()

    # At this point, vertices are positioned in an ECEF frame; transform them to
    # the local ENU frame
    trafo = _transform_lonlat_range_to_local(lon_center, lat_center)
    vertices = _apply(trafo, vertices)
    return vertices


def _transform_lonlat_range_to_local(
    lon_center, lat_center
) -> "mitsuba.ScalarTransform4f":
    """
    Create a transformation matrix to position points placed in ECEF coordinates
    from longitude / latitude information to the local frame in a spherical-shell
    geometry.
    """
    angle_1 = np.rad2deg(lon_center)
    angle_2 = 90.0 - np.rad2deg(lat_center)
    angle_3 = 90.0
    return (
        mi.ScalarTransform4f.rotate(axis=[0, 0, 1], angle=-angle_3)
        @ mi.ScalarTransform4f.rotate(axis=[0, 1, 0], angle=-angle_2)
        @ mi.ScalarTransform4f.rotate(axis=[0, 0, 1], angle=-angle_1)
    )


def triangulate_grid(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray | None = None,
    flip: bool = False,
    divide: t.Literal["nesw", "nwse"] = "nesw",
):
    """
    Create a 2D triangulation for a regular (x, y) grid.

    Parameters
    ----------
    x : ndarray
        List of x grid values.

    y : ndarray
        List of y grid values.

    z : ndarray, optional
        If passed, a 3rd coordinate for vertices in gridded format (y-major).

    flip : bool
        If ``True``, flip triangle orientation (clockwise by default).

    divide : {"nesw", "nwse"}, default: "nesw"
        Cell division method.

    Returns
    -------
    vertices : ndarray
        Vertex list (y-major) as a (n, 2) array.

    faces : ndarray
        Face definitions as a (n, 3) array.
    """
    x_grid, y_grid = np.meshgrid(x, y)
    vertices = np.array((x_grid.ravel(), y_grid.ravel())).T

    xi = np.arange(0, len(x), 1, dtype="int")
    yi = np.arange(0, len(y), 1, dtype="int")

    def _vertex_indices(xg, yg, xstride):
        return xg + xstride * yg

    xg, yg = np.meshgrid(xi[:-1], yi[:-1])
    vertices_sw = _vertex_indices(xg.ravel(), yg.ravel(), xstride=len(x))

    xg, yg = np.meshgrid(xi[1:], yi[:-1])
    vertices_se = _vertex_indices(xg.ravel(), yg.ravel(), xstride=len(x))

    xg, yg = np.meshgrid(xi[:-1], yi[1:])
    vertices_nw = _vertex_indices(xg.ravel(), yg.ravel(), xstride=len(x))

    xg, yg = np.meshgrid(xi[1:], yi[1:])
    vertices_ne = _vertex_indices(xg.ravel(), yg.ravel(), xstride=len(x))

    if divide == "nesw":
        face_indices_1 = np.array((vertices_sw, vertices_se, vertices_ne)).T
        face_indices_2 = np.array((vertices_sw, vertices_ne, vertices_nw)).T
    elif divide == "nwse":
        face_indices_1 = np.array((vertices_sw, vertices_nw, vertices_se)).T
        face_indices_2 = np.array((vertices_nw, vertices_ne, vertices_se)).T
    else:
        raise ValueError(f"unknown cell division method '{divide}'")

    faces = np.concatenate((face_indices_1, face_indices_2))

    if flip:
        faces = faces[:, [0, 2, 1]]

    # If relevant, add elevation as 3rd vertex coordinate
    if z is not None:
        # IMPORTANT: meshgrid() produces an (N_y, N_x)-shaped array, while we
        # expected the z array to be laid out oppositely; hence the transpose
        # prior to flattening
        vertices = np.concatenate((vertices, z.T.ravel().reshape(-1, 1)), axis=1)

    return vertices, faces


def _dem_texcoords(xlon, ylat, xlon_lim, ylat_lim) -> np.ndarray:
    """
    Map the x and y coordinates of the elements of `vertices` to the [0, 1] range
    using a linear scale within xlim and ylim respectively.

    Parameters
    ----------
    xlon : array-like
        An array of shape (N,) that contains the x/longitude coordinates of an
        arbitrary number of vertices.

    ylat : array-like
        An array of shape (N,) that contains the y/latitude coordinates of an
        arbitrary number of vertices.

    xlon_lim : tuple[float, float]
        Lower and upper bounds defining the linear transform that converts the x
        coordinate to the u texture coordinate.

    ylat_lim : tuple[float, float]
        Lower and upper bounds defining the linear transform that converts the y
        coordinate to the v texture coordinate.
    """
    uvs_x = (xlon - xlon_lim[0]) / (xlon_lim[1] - xlon_lim[0])
    uvs_y = (ylat - ylat_lim[0]) / (ylat_lim[1] - ylat_lim[0])

    return np.stack([uvs_x, uvs_y]).T


def mesh_from_dem(
    da: xr.DataArray,
    geometry: str | dict | SceneGeometry,
    planet_radius: pint.Quantity | float | None = None,
    add_texcoords: bool = False,
) -> tuple[BufferMeshShape, pint.Quantity, pint.Quantity]:
    """
    Construct a DEM surface mesh from a data array holding elevation data.

    This function has 4 modes of operations, depending on 2 parameters:

    * the selected geometry type (plane-parallel or spherical-shell);
    * the coordinates of the input dataset (longitude/latitude or x/y).

    Alongside the generated mesh, this function returns extents that can be used
    to position a background stencil around the produced shape (see the notes
    for details).

    Parameters
    ----------
    da : DataArray
        Data array with elevation data, indexed either by latitude and longitude
        coordinates or x and y coordinates.

    geometry : .SceneGeometry or dict or str
        Scene geometry configuration. The value is pre-processed by the
        :meth:`.SceneGeometry.convert` converter.

    planet_radius : quantity or float, default: :data:`.EARTH_RADIUS`
        Planet radius used to convert latitude/longitude to x/y when
        ``geometry`` is a :class:`.PlaneParallelGeometry` instance.
        This parameter is unused otherwise. If a unitless value is passed, it is
        interpreted using
        :ref:`default config length units <sec-user_guide-unit_guide_user>`.

    add_texcoords : bool, default: False
        If ``True``, texture coordinates are added to the created mesh.

    Returns
    -------
    mesh : .BufferMeshShape
        A triangulated mesh representing the DEM.

    xlon_lim : quantity
        Limits of DEM on the x/longitude axis, in length (resp. angle) units for
        plane-parallel (resp. spherical-shell) geometries.

    ylat_lim : quantity
        Limits of DEM on the y/latitude axis, in length (resp. angle) units for
        plane-parallel (resp. spherical-shell) geometries.

    Notes
    -----
    * The ``da`` parameter may use the following formats:

      * with longitude / latitude coordinates, then named ``"lon"`` and
        ``"lat"`` respectively;
      * with x / y coordinates, then named ``"x"`` and ``"y"`` respectively.

      Coordinate and variable units are specified using the ``units`` xarray
      attributes and are expected to be consistent with coordinate names.

    * The mesh generation algorithm operates in four modes, two of which exist
      for testing purposes and not recommended for quantitative applications:

      * **Plane-parallel / xy mode:** The mesh is generated using vertices
        positioned at grid points, then offset to center it at the origin
        location of the local frame. The returned extent is in length units.
      * **Spherical-shell / lonlat mode:** Coordinates are straightforwardly
        converted to an ECEF frame. The resulting mesh is positioned where it
        should fit on the reference geoid (currently a perfect sphere). The
        returned extent is in longitude/latitude.
      * **Plane-parallel / lonlat mode:** The input data is interpreted using a
        Mercator projection, and generated vertices are offset to center the
        mesh at the origin of the local frame. The returned extent is in length
        units. *The projection is highly distorting and should not be used for
        quantitative applications.*
      * **Spherical-shell / xy mode:** Coordinates are converted to longitude
        and latitude using an inverse Mercator projection, assuming that the
        dataset is centred at the equator, then reinterpreted in an ECEF frame.
        The returned extent is in longitude/latitude. *The projection is highly
        distorting and should not be used for quantitative applications.*

    * The generated mesh can optionally be assigned texture coordinates used to
      map spatially-varying data (*e.g.* textured reflectance value for a
      Lambertian BSDF).
    """
    # Pre-process geometry parameter
    geometry = SceneGeometry.convert(geometry)

    if planet_radius is not None:
        if isinstance(geometry, SphericalShellGeometry):
            warnings.warn(
                "The ``planet_radius`` argument is set to a user-defined "
                "value but will be overridden by the value held by the "
                "SphericalGeometry instance passed as the ``geometry`` "
                "parameter."
            )

        # Add default units if quantity is unitless
        planet_radius = pinttrs.util.ensure_units(planet_radius, ucc.get("length"))

    # Set default planet radius value
    planet_radius = EARTH_RADIUS if planet_radius is None else planet_radius
    if isinstance(geometry, SphericalShellGeometry):
        planet_radius = geometry.planet_radius

    # Check data array coordinates and dimensions
    length_kernel_u = uck.get("length")

    if ("lon" in da.coords) and ("lat" in da.coords):
        mode = "lonlat"
        xlon_dim, ylat_dim = "lon", "lat"

    elif ("x" in da.coords) and ("y" in da.coords):
        mode = "xy"
        xlon_dim, ylat_dim = "x", "y"

    else:
        raise ValueError(
            "Data array coordinates must include either `x/y` or `lon/lat`.\n"
            f"Got: {da.coords}"
        )

    # Convert to distances in plane-parallel geometry,
    # and to angles in spherical-shell geometry.
    xlon = to_quantity(da[xlon_dim])
    ylat = to_quantity(da[ylat_dim])
    xlon_center = _middle(xlon)
    ylat_center = _middle(ylat)
    # Extract elevation data and ensure y-major layout
    elevation = to_quantity(da.transpose(xlon_dim, ylat_dim))
    # By default, no texture coordinates are assigned
    texcoords = None

    # Process vertex data depending on geometry and coordinate mode, generate triangulation
    if isinstance(geometry, PlaneParallelGeometry):
        if mode == "xy":
            xlon = xlon - xlon_center
            ylat = ylat - ylat_center

            vertices, faces = triangulate_grid(
                xlon.m_as(length_kernel_u),
                ylat.m_as(length_kernel_u),
                elevation.m_as(length_kernel_u),
            )

            xlon_lim = (xlon.m.min(), xlon.m.max()) * xlon.u
            ylat_lim = (ylat.m.min(), ylat.m.max()) * ylat.u

            # If relevant, assign texture coordinates
            if add_texcoords:
                texcoords = _dem_texcoords(
                    vertices[:, 0], vertices[:, 1], xlon_lim.m, ylat_lim.m
                )

        elif mode == "lonlat":
            x, y = _mercator(
                xlon.m_as(ureg.rad),
                ylat.m_as(ureg.rad),
                planet_radius.m_as(length_kernel_u),
            )
            da = (
                da.assign_coords(
                    {
                        "x": ("lon", x, {"units": symbol(length_kernel_u)}),
                        "y": ("lat", y, {"units": symbol(length_kernel_u)}),
                    }
                )
                .swap_dims({"lon": "x", "lat": "y"})
                .drop_vars(("lon", "lat"))
            )
            return mesh_from_dem(
                da, geometry, planet_radius, add_texcoords=add_texcoords
            )

        else:
            raise RuntimeError(f"unknown input mode {mode}")

    elif isinstance(geometry, SphericalShellGeometry):
        if mode == "xy":
            lon, lat = _mercator_inverse(
                xlon.m_as(length_kernel_u),
                ylat.m_as(length_kernel_u),
                planet_radius.m_as(length_kernel_u),
            )
            da = (
                da.assign_coords(
                    {
                        "lon": ("x", np.rad2deg(lon), {"units": "degree"}),
                        "lat": ("y", np.rad2deg(lat), {"units": "degree"}),
                    }
                )
                .swap_dims(({"x": "lon", "y": "lat"}))
                .drop_vars(("x", "y"))
            )
            return mesh_from_dem(
                da, geometry, planet_radius, add_texcoords=add_texcoords
            )

        elif mode == "lonlat":
            xlon = xlon.to(ureg.rad)
            ylat = ylat.to(ureg.rad)
            vertices, faces = triangulate_grid(
                xlon.m, ylat.m, elevation.m_as(length_kernel_u)
            )

            # If relevant, assign texture coordinates
            xlon_lim = (xlon.m.min(), xlon.m.max()) * xlon.u
            ylat_lim = (ylat.m.min(), ylat.m.max()) * ylat.u
            if add_texcoords:
                texcoords = _dem_texcoords(
                    vertices[:, 0], vertices[:, 1], xlon_lim.m, ylat_lim.m
                )

            # Rotate mesh to Eradiate's local frame (located at the North pole)
            vertices = _transform_vertices_spherical_shell_lonlat(
                vertices, planet_radius.m_as(length_kernel_u)
            )
            # Recompute limits (returned afterwards)
            xlon_lim = (xlon.m.min(), xlon.m.max()) * xlon.u
            ylat_lim = (ylat.m.min(), ylat.m.max()) * ylat.u

        else:
            raise RuntimeError(f"unknown input mode {mode}")

    else:  # For completeness
        raise TypeError(f"unhandled geometry type '{type(PlaneParallelGeometry)}'")

    # Create mesh instance
    mesh = BufferMeshShape(vertices=vertices, faces=faces, texcoords=texcoords)

    return mesh, xlon_lim, ylat_lim


@define(eq=False, slots=False)
class DEMSurface(Surface):
    """
    DEM Surface [``dem``]

    A mesh-based representation of a Digital Elevation Model (DEM). This class
    holds a mesh shape instance, as well as a background shape that provides a
    surface beyond the extents of the mesh.

    The intended instantiation method is to, first, create a mesh using the
    :func:`.mesh_from_dem`, then use the data it returns to call the
    :meth:`.from_mesh` constructor.
    """

    id: str | None = documented(
        attrs.field(
            default="terrain",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(Surface, "id", "doc"),
        type=get_doc(Surface, "id", "type"),
        init_type=get_doc(Surface, "id", "init_type"),
        default='"surface"',
    )

    shape: Shape = documented(
        attrs.field(
            converter=attrs.converters.optional(shape_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of((BufferMeshShape, FileMeshShape))
            ),
            kw_only=True,
            default=None,
        ),
        doc="Shape describing the surface.",
        type=".BufferMeshShape or .FileMeshShape or None",
        init_type=".BufferMeshShape or .FileMeshShape or dict",
        default="None",
    )

    shape_background: Shape = documented(
        attrs.field(
            converter=attrs.converters.optional(shape_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of((SphereShape, RectangleShape))
            ),
            kw_only=True,
            default=None,
        ),
        doc="Shape describing the background surface.",
        type=".SphereShape or .RectangleShape or None",
        init_type=".SphereShape or .RectangleShape or dict, optional",
        default="None",
    )

    @property
    def _shape_id(self) -> str:
        """
        Mitsuba shape object identifier.
        """
        return f"{self.id}_shape"

    @property
    def _bsdf_id(self) -> str:
        """
        Mitsuba BSDF object identifier.
        """
        return f"{self.id}_bsdf"

    @property
    def _template_shapes(self) -> dict:
        return {}

    @property
    def _params_shapes(self) -> dict:
        return {}

    @property
    def _template_bsdfs(self) -> dict:
        return {}

    @property
    def _params_bsdfs(self) -> dict:
        return {}

    @property
    def objects(self) -> dict[str, SceneElement]:
        # Inherit docstring
        return {
            self._shape_id: self.shape,
            f"{self._shape_id}_background": self.shape_background,
        }

    @classmethod
    def from_mesh(
        cls,
        mesh: BufferMeshShape | FileMeshShape,
        xlon_lim: pint.Quantity,
        ylat_lim: pint.Quantity,
        id: str = "surface",
        geometry: SceneGeometry | str | dict = "plane_parallel",
        planet_radius: pint.Quantity | None = None,
        bsdf: BSDF | None = None,
        bsdf_mesh: BSDF | None = None,
        bsdf_background: BSDF | None = None,
    ) -> DEMSurface:
        """
        Construct a :class:`.DEMSurface` instance from a mesh object and
        coordinates which specify its location.

        Parameters
        ----------
        mesh : .BufferMeshShape or .FileMeshShape
            DEM as a triangulated mesh. The BSDF of this shape will be
            overridden by the ``bsdf_mesh`` parameter.

        xlon_lim : quantity
            Limits of the x/longitude range covered by the mesh.

        ylat_lim : quantity
            Limits of the y/latitude range covered by the mesh.

        id : str, default: "surface"
            Identifier of the scene element.

        geometry : .SceneGeometry or str or dict, default: "plane_parallel"
            Scene geometry. Strings and dictionaries are processed by the
            :meth:`.SceneGeometry.convert` function.

        planet_radius : quantity, default: .EARTH_RADIUS
            Planet radius. Used only in case of a plane parallel geometry to
            convert between latitude/longitude and x/y coordinates.

        bsdf : .BSDF, default: :class:`LambertianBSDF() <.LambertianBSDF>`
            Alias to ``bsdf_mesh`` (**deprecated**). If both are defined,
            ``bsdf_mesh`` takes precedence.

        bsdf_mesh : .BSDF, default: :class:`LambertianBSDF() <.LambertianBSDF>`
            Scattering model attached to the mesh.

        bsdf_background : .BSDF, default: :class:`LambertianBSDF() <.LambertianBSDF>`
            Scattering model attached to the background shape.

        Returns
        -------
        .DEMSurface
        """
        geometry = SceneGeometry.convert(geometry)

        if bsdf_mesh is None:
            bsdf_mesh = LambertianBSDF() if bsdf is None else bsdf

        else:
            if bsdf is not None:
                warnings.warn(
                    "while calling DEMSurface.from_mesh(): "
                    "both bsdf and bsdf_mesh parameters were specified; "
                    "bsdf_mesh takes precedence"
                )

        bsdf_background = (
            LambertianBSDF() if bsdf_background is None else bsdf_background
        )

        mesh = attrs.evolve(mesh, bsdf=bsdf_mesh)

        if isinstance(geometry, PlaneParallelGeometry):
            kernel_length_units = uck.get("length")
            g_width = geometry.width.to(kernel_length_units)
            g_altitude = geometry.ground_altitude.to(kernel_length_units)

            x_range = (xlon_lim[1] - xlon_lim[0]).m_as(kernel_length_units)
            x_scale = g_width.m / (x_range * 3.0)

            y_range = (ylat_lim[1] - ylat_lim[0]).m_as(kernel_length_units)
            y_scale = g_width.m / (y_range * 3.0)

            opacity_array = np.array(
                [
                    [1.0, 1.0, 1.0],
                    [1.0, 0.0, 1.0],
                    [1.0, 1.0, 1.0],
                ],
            )
            opacity_mask_trafo = mi.ScalarTransform4f.scale(
                [x_scale, y_scale, 1.0]
            ) @ mi.ScalarTransform4f.translate(
                [-0.5 + (0.5 / x_scale), -0.5 + (0.5 / y_scale), 0.0]
            )
            opacity_bsdf = OpacityMaskBSDF(
                nested_bsdf=bsdf_background,
                opacity_bitmap=opacity_array,
                uv_trafo=opacity_mask_trafo,
            )
            surface_background = RectangleShape(
                id=f"{id}_background",
                edges=g_width,
                center=[0.0, 0.0, g_altitude.m] * g_altitude.units,
                normal=[0, 0, 1],
                up=[0, 1, 0],
                bsdf=opacity_bsdf,
            )

        elif isinstance(geometry, SphericalShellGeometry):
            if planet_radius is not None:
                warnings.warn(
                    "SphericalShellGeometry overrides the `planet_radius` argument."
                )

            def _to_uv(lon_lim, lat_lim) -> "mitsuba.ScalarTransform4f":
                """
                Compute the `to_uv` transformation for the opacity mask bitmap.
                It moves the central (transparent) part of the bitmap to where
                it covers the specified longitude/latitude extent.
                """
                lon_range = lon_lim[1] - lon_lim[0]
                lon_scale = 120.0 / lon_range

                lat_range = lat_lim[1] - lat_lim[0]
                lat_scale = 60.0 / lat_range

                lon_middle = _middle(lon_lim)
                lon_uv = lon_middle / 360.0 + 0.5

                lat_mean = (lat_lim[1] + lat_lim[0]) / 2.0
                lat_uv = 0.5 - (lat_mean / 180)

                return mi.ScalarTransform4f.scale(
                    (lon_scale, lat_scale, 1.0)
                ) @ mi.ScalarTransform4f.translate(
                    [-lon_uv + (0.5 / lon_scale), -lat_uv + (0.5 / lat_scale), 0.0]
                )

            opacity_mask_trafo = _to_uv(
                xlon_lim.m_as(ureg.deg), ylat_lim.m_as(ureg.deg)
            )
            opacity_array = np.array(
                [
                    [1.0, 1.0, 1.0],
                    [1.0, 0.0, 1.0],
                    [1.0, 1.0, 1.0],
                ],
            )
            opacity_bsdf = OpacityMaskBSDF(
                nested_bsdf=bsdf_background,
                opacity_bitmap=opacity_array,
                uv_trafo=opacity_mask_trafo,
            )

            # The 'hole' in the background surface is created at the equator:
            # rotate the shape to align it with the mesh
            trafo = (
                mi.ScalarTransform4f.rotate(axis=[0, 0, 1], angle=90)
                @ mi.ScalarTransform4f.rotate(
                    axis=[0, 1, 0], angle=(90 - _middle(ylat_lim.m_as(ureg.deg)))
                )
                @ mi.ScalarTransform4f.rotate(
                    axis=[0, 0, 1], angle=-_middle(xlon_lim.m_as(ureg.deg))
                )
            )

            surface_background = SphereShape(
                id=f"{id}_shape_background",
                center=[0.0, 0.0, 0.0] * geometry.planet_radius.units,
                radius=geometry.planet_radius + geometry.ground_altitude,
                bsdf=opacity_bsdf,
                to_world=trafo,
            )

        else:
            raise TypeError(f"Unhandled scene geometry type '{type(geometry)}'")

        return cls(shape=mesh, shape_background=surface_background, id=id)



from __future__ import annotations

import itertools
from enum import Enum
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urljoin

import pooch

from ._canopy_loader import load_scenario

_DATA_URL_ROOT = "https://eradiate.eu/data/store/unstable/scenarios/rami5/"


class RAMIActualCanopies(Enum):
    """Enumeration of RAMI actual canopies."""

    JARVSELJA_PINE_STAND = "HET07_JPS_SUM"  # Summer
    OFENPASS_PINE_STAND = "HET08_OPS_WIN"  # Winter
    JARVSELJA_BIRCH_STAND_SUMMER = "HET09_JBS_SUM"  # Summer
    WELLINGTON_CITRUS_ORCHARD = "HET14_WCO_UND"
    JARVSELJA_BIRCH_STAND_WINTER = "HET15_JBS_WIN"  # Winter
    AGRICULTURAL_CROPS = "HET16_SRF_UND"  # Short Rotation Forest
    SAVANNA_PRE_FIRE = "HET50_SAV_PRE"  # Semi-empirical
    WYTHAM_WOOD = "HET51_WWO_TLS"  # Empirical


class RAMIHeterogeneousAbstractCanopies(Enum):
    """Enumeration of RAMI heterogeneous abstract canopies."""

    ANISOTROPIC_BACKGROUND_OVERSTOREY_SPARSE_BRF_MODEL_A = "HET10_DIS_S1A"
    ANISOTROPIC_BACKGROUND_OVERSTOREY_SPARSE_BRF_MODEL_B = "HET11_DIS_S1B"
    ANISOTROPIC_BACKGROUND_OVERSTOREY_SPARSE_BRF_MODEL_C = "HET12_DIS_S1C"
    ANISOTROPIC_BACKGROUND_OVERSTOREY_DENSE_BRF_MODEL_A = "HET20_DIS_D1A"
    ANISOTROPIC_BACKGROUND_OVERSTOREY_DENSE_BRF_MODEL_B = "HET21_DIS_D1B"
    ANISOTROPIC_BACKGROUND_OVERSTOREY_DENSE_BRF_MODEL_C = "HET22_DIS_D1C"
    TWO_LAYER_CANOPY_OVERSTORIES_SPARSE_UNDERSTORIES_SPARSE = "HET16_DIS_S2S"
    TWO_LAYER_CANOPY_OVERSTORIES_MEDIUM_UNDERSTORIES_SPARSE = "HET17_DIS_M2S"
    TWO_LAYER_CANOPY_OVERSTORIES_DENSE_UNDERSTORIES_SPARSE = "HET18_DIS_D2S"
    TWO_LAYER_CANOPY_OVERSTORIES_SPARSE_UNDERSTORIES_DENSE = "HET26_DIS_S2D"
    TWO_LAYER_CANOPY_OVERSTORIES_MEDIUM_UNDERSTORIES_DENSE = "HET27_DIS_M2D"
    TWO_LAYER_CANOPY_OVERSTORIES_DENSE_UNDERSTORIES_DENSE = "HET28_DIS_D2D"
    CONSTANT_SLOPE_DISTRIBUTION_SPARSE_INCLINATION_15 = "HET23_DIS_S15"
    CONSTANT_SLOPE_DISTRIBUTION_DENSE_INCLINATION_15 = "HET24_DIS_D15"
    CONSTANT_SLOPE_DISTRIBUTION_SPARSE_INCLINATION_30 = "HET33_DIS_S30"
    CONSTANT_SLOPE_DISTRIBUTION_DENSE_INCLINATION_30 = "HET34_DIS_D30"


class RAMIHomogeneousAbstractCanopies(Enum):
    """Enumeration of RAMI homogeneous abstract canopies."""

    ANISOTROPIC_BACKGROUND_PLANOPHILE_A = "HOM23_DIS_P1A"
    ANISOTROPIC_BACKGROUND_PLANOPHILE_B = "HOM24_DIS_P1B"
    ANISOTROPIC_BACKGROUND_PLANOPHILE_C = "HOM25_DIS_P1C"
    # ANISOTROPIC_BACKGROUND_ERECTOPHILE_A = "HOM33_DIS_E1A"
    ANISOTROPIC_BACKGROUND_ERECTOPHILE_B = "HOM34_DIS_E1B"
    ANISOTROPIC_BACKGROUND_ERECTOPHILE_C = "HOM35_DIS_E1C"
    TWO_LAYER_CANOPY_ERECTOPHILE_SPARSE_PLANOPHILE_DENSE = "HOM26_DIS_EPD"
    TWO_LAYER_CANOPY_ERECTOPHILE_SPARSE_PLANOPHILE_MEDIUM = "HOM27_DIS_EPM"
    TWO_LAYER_CANOPY_ERECTOPHILE_SPARSE_PLANOPHILE_SPARSE = "HOM28_DIS_EPS"
    TWO_LAYER_CANOPY_PLANOPHILE_SPARSE_ERECTOPHILE_DENSE = "HOM36_DIS_PED"
    TWO_LAYER_CANOPY_PLANOPHILE_SPARSE_ERECTOPHILE_MEDIUM = "HOM37_DIS_PEM"
    TWO_LAYER_CANOPY_PLANOPHILE_SPARSE_ERECTOPHILE_SPARSE = "HOM38_DIS_PES"
    ADJACENT_CANOPIES_SPARSE_ERECTOPHILE_DENSE_PLANOPHILE = "HOM29_DIS_EM0"
    ADJACENT_CANOPIES_MEDIUM_ERECTOPHILE_SPARSE_PLANOPHILE = "HOM30_DIS_ED0"


RAMICanopies = Union[
    RAMIActualCanopies,
    RAMIHomogeneousAbstractCanopies,
    RAMIHeterogeneousAbstractCanopies,
]  #: Type alias to a union of all RAMI canopy enumerations.


class RAMIScenarioVariant(Enum):
    """Enumeration of RAMI scenario variants (original or simplified)."""

    ORIGINAL = "original"
    SIMPLIFIED = "simplified"


def generate_name(
    scenario_name: RAMICanopies,
    variant: RAMIScenarioVariant = RAMIScenarioVariant.ORIGINAL,
) -> str:
    """
    Generate a name for a scenario based on its name and variant.

    Parameters
    ----------
    scenario_name : RAMIActualCanopies or RAMIHeterogeneousAbstractCanopies or RAMIHomogeneousAbstractCanopies
        The name of the scenario.
    variant : ScenarioVersion
        The variant of the scenario.

    Returns
    -------
    str
        The name of the scenario.
    """
    return (
        f"{scenario_name.value}-{variant.value}"
        if variant == RAMIScenarioVariant.SIMPLIFIED
        else scenario_name.value
    )


def _convert_to_enum(scenario_name: str | RAMICanopies) -> RAMICanopies:
    """
    Convert a scenario name to an enum if it is a string.

    Parameters
    ----------
    scenario_name : str or RAMICanopies
        The name of the scenario.

    Returns
    -------
    (RAMIActualCanopies | RAMIHeterogeneousAbstractCanopies | RAMIHomogeneousAbstractCanopies)
        The name of the scenario as an enum.
    """

    if isinstance(scenario_name, str):
        for member in itertools.chain.from_iterable(
            [
                RAMIActualCanopies,
                RAMIHeterogeneousAbstractCanopies,
                RAMIHomogeneousAbstractCanopies,
            ]
        ):
            if scenario_name == member.value:
                return member
        else:
            raise ValueError(f"Scenario {scenario_name} not found")
    else:
        return scenario_name


def load_rami_scenario(
    scenario_name: str | RAMICanopies,
    variant: RAMIScenarioVariant = RAMIScenarioVariant.ORIGINAL,
    padding: int = 0,
    unpack_folder: Optional[Path] = None,
    spectral_data: Optional[dict] = None,
) -> dict:
    """
    Load a scenario based on its name and variant.

    This function will check if scenario data can be found at the target
    location; if not, it will download them automatically.

    Parameters
    ----------
    scenario_name : str or RAMIActualCanopies or RAMIHeterogeneousAbstractCanopies or RAMIHomogeneousAbstractCanopies
        The name of the RAMI-V scenario. If a string is provided, it will
        automatically be converted to the appropriate enum.
    variant : RAMIScenarioVariant
        The variant of the scenario.
    padding : int, optional
        The padding to apply to the scenario, defaults to 0.
    unpack_folder : path-like, optional
        Directory where scenario data is expected to be stored — and where
        downloaded data will be unpacked. Defaults to ``$PWD``.
    spectral_data : dict[str, Any or dict[str, Any]] or None
        Spectral data to apply to the scenario, defaults to None (keep original).

    Returns
    -------
    dict
        The scenario.

    See Also
    --------
    load_scenario
    """
    unpack_folder = Path.cwd() if unpack_folder is None else Path(unpack_folder)
    name = generate_name(_convert_to_enum(scenario_name), variant)
    fname = f"{name}.zip"

    # Check for data availability
    scenario_folder = unpack_folder / name
    if not (scenario_folder / "scenario.json").exists():
        pooch.retrieve(
            urljoin(_DATA_URL_ROOT, fname),
            fname=fname,
            path=unpack_folder,
            processor=pooch.processors.Unzip(extract_dir=name),
            known_hash=None,
            progressbar=True,
        )
        (unpack_folder / fname).unlink(missing_ok=True)

    # Load the scenario
    return load_scenario(scenario_folder, padding, spectral_data=spectral_data)



"""Utility components for quadrature rules."""

from __future__ import annotations

from enum import Enum

import attrs
import numpy as np

from .attrs import define, documented
from .util.misc import str_summary_numpy


class QuadType(Enum):
    """Quadrature rule type flags."""

    GAUSS_LEGENDRE = "gauss_legendre"
    GAUSS_LOBATTO = "gauss_lobatto"


@define(eq=False, frozen=True)
class Quad:
    """
    A data class storing information about a quadrature rule. Nodes and weights
    are defined in the [-1, 1] interval. The reference interval can be changed
    using the ``interval`` argument of the :meth:`.eval_nodes` and
    :meth:`.integrate` functions.
    """

    type: QuadType = documented(
        attrs.field(converter=QuadType, repr=lambda x: str(x)),
        doc="Quadrature type. If a string is passed, it is converted to a "
        ":class:`.QuadType`.",
        type=":class:`.QuadType`",
    )

    nodes: np.ndarray = documented(
        attrs.field(converter=np.array, repr=str_summary_numpy),
        doc="Quadrature rule nodes.",
        type="ndarray",
    )

    weights: np.ndarray = documented(
        attrs.field(converter=np.array, repr=str_summary_numpy),
        doc="Quadrature rule weights.",
        type="ndarray",
    )

    @nodes.validator
    @weights.validator
    def _nodes_weights_validator(self, attribute, value):
        if self.nodes.shape != self.weights.shape:
            raise ValueError(
                f"while validating {attribute.name}: nodes and weights arrays "
                f"must have the same shape, got nodes.shape = {self.nodes.shape} "
                f"and weights.shape = {self.weights.shape}"
            )

    def pretty_repr(self) -> str:
        return f"{self.type.value}, {self.nodes.size} points"

    @classmethod
    def gauss_legendre(cls, n: int) -> Quad:
        """
        Initialize a :class:`.Quad` instance with Gauss-Legendre nodes and
        weights.

        Parameters
        ----------
        n : int
            Number of quadrature points.

        Returns
        -------
        :class:`.Quad`
            Gauss-Legendre quadrature definition.
        """
        from mitsuba.scalar_rgb.quad import gauss_legendre

        nodes, weights = gauss_legendre(n)
        return cls(
            type=QuadType.GAUSS_LEGENDRE,
            nodes=np.array(nodes, dtype=float),
            weights=np.array(weights, dtype=float),
        )

    @classmethod
    def gauss_lobatto(cls, n: int) -> Quad:
        """
        Initialize a :class:`.Quad` instance with Gauss-Lobatto nodes and
        weights.

        Parameters
        ----------
        n : int
            Number of quadrature points.

        Returns
        -------
        :class:`.Quad`
            Gauss-Lobatto quadrature definition.
        """
        from mitsuba.scalar_rgb.quad import gauss_lobatto

        nodes, weights = gauss_lobatto(n)
        return cls(
            type=QuadType.GAUSS_LOBATTO,
            nodes=np.array(nodes, dtype=float),
            weights=np.array(weights, dtype=float),
        )

    @classmethod
    def new(cls, type: str | QuadType, n: int) -> Quad:
        """
        Initialize a :class:`.Quad` instance of the specified type.

        Parameters
        ----------
        type : str or .QuadType
            Quadrature rule type. If a string is passed, it is converted to
            a :class:`.QuadType`.

        n : int
            Number of quadrature points.

        Returns
        -------
        :class:`.Quad`
            Quadrature definition.
        """
        type = QuadType(type)

        if type is QuadType.GAUSS_LEGENDRE:
            return cls.gauss_legendre(n)

        elif type is QuadType.GAUSS_LOBATTO:
            return cls.gauss_lobatto(n)

        else:
            raise ValueError(f"unknown quadrature type '{type}'")

    def eval_nodes(
        self, interval: tuple[float, float] | None = None
    ) -> np.typing.ArrayLike:
        """
        Compute nodes scaled to a specific interval.

        Parameters
        ----------
        interval :  tuple of float, optional
            Interval for which nodes are to be scaled as a 2-tuple. If ``None``,
            the default [-1, 1] is used.

        Returns
        -------
        ndarray
            Scaled node values.
        """
        if interval is None:
            return self.nodes
        a, b = interval
        return 0.5 * (a + b + (b - a) * self.nodes)

    def integrate(
        self, values: np.typing.ArrayLike, interval: tuple[float, float] | None
    ) -> float:
        """
        Evaluate quadrature rule, accounting for interval scaling.

        Parameters
        ----------
        values : ndarray
            Function values at quadrature nodes.

        interval : tuple of float, optional
            Interval on which the integral is being computed as a 2-tuple.
            If ``None``, the default [-1, 1] is used.

        Returns
        -------
        float
            Quadrature evaluation for the specified interval.
        """

        weighted_sum = float(np.dot(self.weights, values))

        if interval is None:
            return weighted_sum
        else:
            return 0.5 * (interval[1] - interval[0]) * weighted_sum

    @property
    def str_summary(self) -> str:
        """
        Return a summarized representation of the current instance.

        Returns
        -------
        str
            Instance summary.
        """
        return f"Quad(type={QuadType(self.type)}, n={len(self.nodes)})"



from __future__ import annotations

import attrs
import mitsuba as mi
import pint
import pinttr

from ._core import BSDF
from ..core import traverse
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import define, documented
from ...kernel import SceneParameter, SearchSceneParameter
from ...units import unit_context_config as ucc


@define(eq=False, slots=False)
class RTLSBSDF(BSDF):
    """
    RTLS BSDF [``rtls``].

    This class implements the RossThick-LiSparse (RTLS) BRDF as described
    by the MODIS BRDF/Albedo Product ATBD :cite:`BU-MODISBRDFAlbedoProductATBD1999`.

    See Also
    --------
    :ref:`plugin-bsdf-rtls`
    """

    f_iso: Spectrum = documented(
        attrs.field(
            default=0.209741,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Isotropic scattering kernel parameter. "
        r"Defaults to :math:`f_{iso} = 0.209741`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float, optional",
        default="0.209741",
    )

    f_vol: Spectrum = documented(
        attrs.field(
            default=0.004140,
            converter=spectrum_factory.converter("dimensionless"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                validators.has_quantity("dimensionless"),
            ],
        ),
        doc="Volumetric scattering from horizontally homogeneous leaf canopies "
        r"kernel parameter. Defaults to :math:`f_{vol} = 0.004140`.",
        type=".Spectrum",
        init_type=".Spectrum or dict or float, optional",
        default="0.004140",
    )

    f_geo: Spectrum | None = documented(
        attrs.field(
            default=0.081384,
            converter=attrs.converters.optional(
                spectrum_factory.converter("dimensionless")
            ),
            validator=attrs.validators.optional(
                [
                    attrs.validators.instance_of(Spectrum),
                    validators.has_quantity("dimensionless"),
                ]
            ),
        ),
        doc="Geometric-optical surface scattering kernel parameter. "
        r"Defaults to :math:`f_{geo} = 0.081384`.",
        type="float",
        init_type="float, optional",
        default="0.081384",
    )

    h: pint.Quantity = documented(
        pinttr.field(
            default=2.0,
            units=ucc.deferred("dimensionless"),
        ),
        doc="Height-to-center-of-crown. Must be dimensionless.",
        type="quantity",
        init_type="quantity or float",
        default="2.0",
    )

    r: pint.Quantity = documented(
        pinttr.field(
            default=1.0,
            units=ucc.deferred("dimensionless"),
        ),
        doc="Crown horizontal radius. Must not be zero.",
        type="quantity",
        init_type="quantity or float",
        default="1.0",
    )

    b: pint.Quantity = documented(
        pinttr.field(
            default=1.0,
            units=ucc.deferred("dimensionless"),
        ),
        doc="Crown vertical radius. Must not be zero.",
        type="quantity",
        init_type="quantity or float",
        default="1.0",
    )

    @r.validator
    def _r_validator(self, attribute, value):
        assert value != 0.0

    @b.validator
    def _b_validator(self, attribute, value):
        assert value != 0.0

    @property
    def template(self) -> dict:
        # Inherit docstring
        objects = {
            "f_iso": traverse(self.f_iso)[0],
            "f_vol": traverse(self.f_vol)[0],
            "f_geo": traverse(self.f_geo)[0],
        }

        result = {
            "type": "rtls",
            "h": self.h.magnitude,
            "r": self.r.magnitude,
            "b": self.b.magnitude,
        }

        for obj_key, obj_values in objects.items():
            for key, value in obj_values.items():
                result[f"{obj_key}.{key}"] = value

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        objects = {
            "f_iso": traverse(self.f_iso)[1],
            "f_vol": traverse(self.f_vol)[1],
            "f_geo": traverse(self.f_geo)[1],
        }

        result = {}
        for obj_key, obj_params in objects.items():
            for key, param in obj_params.items():
                result[f"{obj_key}.{key}"] = attrs.evolve(
                    param,
                    search=SearchSceneParameter(
                        node_type=mi.BSDF,
                        node_id=self.id,
                        parameter_relpath=f"{obj_key}.{key}",
                    )
                    if self.id is not None
                    else None,
                )

        return result



from __future__ import annotations

import attrs
import pint

from ._core import Spectrum
from ...attrs import define
from ...kernel import DictParameter, KernelSceneParameterFlags, SceneParameter
from ...radprops.rayleigh import compute_sigma_s_air
from ...units import PhysicalQuantity
from ...units import unit_context_kernel as uck


@define(eq=False, slots=False)
class AirScatteringCoefficientSpectrum(Spectrum):
    """
    Air scattering coefficient spectrum [``air_scattering_coefficient``].

    See Also
    --------
    :func:`~eradiate.radprops.rayleigh.compute_sigma_s_air`
    """

    quantity: PhysicalQuantity = attrs.field(
        default=PhysicalQuantity.COLLISION_COEFFICIENT,
        init=False,
        repr=False,
    )

    def eval_mono(self, w: pint.Quantity) -> pint.Quantity:
        # Inherit docstring
        return compute_sigma_s_air(wavelength=w)

    def eval_ckd(self, w: pint.Quantity, g: float) -> pint.Quantity:
        # Inherit docstring
        return self.eval_mono(w)

    def integral(self, wmin: pint.Quantity, wmax: pint.Quantity) -> pint.Quantity:
        raise NotImplementedError

    @property
    def template(self) -> dict:
        # Inherit docstring

        return {
            "type": "uniform",
            "value": DictParameter(
                func=lambda ctx: float(
                    self.eval(ctx.si).m_as(uck.get("collision_coefficient"))
                )
            ),
        }

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring

        return {
            "value": SceneParameter(
                func=lambda ctx: float(
                    self.eval(ctx.si).m_as(uck.get("collision_coefficient"))
                ),
                flags=KernelSceneParameterFlags.SPECTRAL,
            )
        }



from __future__ import annotations

import attrs
import mitsuba as mi
import numpy as np
import xarray as xr

from ._core import BSDF
from ... import converters
from ...attrs import define, documented
from ...kernel import DictParameter, SceneParameter
from ...units import to_quantity
from ...units import unit_registry as ureg
from ...util.misc import summary_repr


@define(eq=False, slots=False)
class MQDiffuseBSDF(BSDF):
    """
    Measured quasi-diffuse BSDF [``mqdiffuse``].

    This BSDF models the reflection of light by opaque materials with a
    behaviour close to diffuse, *i.e* with no strong scattering lobe.
    Assumptions are as follows:

    * The material is isotropic. Consequently, only the azimuth difference
      matters.
    * The material is gray. Consequently, no spectral dimension is used.

    Notes
    -----
    The input is specified using an xarray dataset. It must contain a ``brdf``
    data variable with the following dimensions (the corresponding coordinate
    range is specified within brackets):

    * ``cos_theta_o`` [0, 1]: cosine of the outgoing zenith angle;
    * ``phi_d`` [0, 2π[: difference between the incoming and outgoing azimuth
      angles;
    * ``cos_theta_i`` [0, 1]: cosine of the incoming zenith angle.

    Coordinates must be evenly spaced and have a `"units"` metadata field.

    Warnings
    --------
    * Table values are not checked internally: ensuring that the data is
      consistent (*e.g* that the corresponding reflectance is not greater than
      1) is the user's responsibility.
    * While this BSDF may technically represent any isotropic material, its
      sampling routine's performance degrades as the material departs from a
      diffuse behaviour.
    """

    data: xr.Dataset = documented(
        attrs.field(
            converter=converters.passthrough_type(xr.Dataset)(converters.load_dataset),
            kw_only=True,
            repr=summary_repr,
        ),
        type="Dataset",
        init_type="Dataset",
        doc="Measured quasi-diffuse BRDF data formatted as an xarray dataset.",
    )

    @data.validator
    def _data_validator(self, attribute, value):
        # Check type
        attrs.validators.instance_of(xr.Dataset)(self, attribute, value)

        # Check data variable
        if "brdf" not in value.data_vars:
            raise ValueError(
                f"while validating '{attribute.name}': missing required data "
                "variable 'brdf'"
            )

        # Check dimensions
        if set(value.data_vars["brdf"].dims) != {"cos_theta_o", "phi_d", "cos_theta_i"}:
            raise ValueError(
                f"while validating '{attribute.name}': incorrect dimension "
                f"list (got '{set(value.dims)}', "
                f"expected '{ {'cos_theta_o', 'phi_d', 'cos_theta_i'} }')"
            )

        # Check coordinates
        for coord_name in ["cos_theta_o", "cos_theta_i", "phi_d"]:
            try:
                coord = to_quantity(value.coords[coord_name])
            except ValueError as e:
                if e.args[0] == "this DataArray has no 'units' metadata field":
                    raise ValueError(
                        f"while validating '{attribute.name}': input dataset "
                        f"coordinate variable '{coord_name}' is missing "
                        "'units' metadata field"
                    ) from e
                else:
                    raise e

            expected = (
                np.linspace(0.0, 1.0, len(coord))
                if coord_name.startswith("cos")
                else (
                    np.linspace(0.0, 2.0 * np.pi, len(coord), endpoint=False) * ureg.rad
                )
            )
            if not np.allclose(coord, expected):
                raise ValueError(
                    f"while validating '{attribute.name}': incorrect "
                    f"coordinate values for field '{coord_name}'; got {coord}, "
                    f"expected {expected}"
                )

    def _eval_grid_impl(self, ctx):
        values = (
            self.data.data_vars["brdf"]
            .transpose("cos_theta_o", "phi_d", "cos_theta_i")
            .values
        )
        # Add an extra row with φ_d = 0° data to ensure azimuthal periodicity
        values = np.concatenate((values, values[:, [0], :]), axis=1)
        return mi.VolumeGrid(values.astype(np.float32))

    @property
    def template(self) -> dict:
        # Inherit docstring

        result = {
            "type": "mqdiffuse",
            "grid": DictParameter(lambda ctx: self._eval_grid_impl(ctx)),
        }

        if self.id is not None:
            result["id"] = self.id

        return result

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring
        return {}



from __future__ import annotations

import typing as t

import attrs

from .attrs import define, documented
from .spectral import SpectralIndex

# ------------------------------------------------------------------------------
#                                      ABC
# ------------------------------------------------------------------------------


@attrs.define
class Context:
    """Abstract base class for all context data structures."""

    def evolve(self, **changes):
        """
        Create a copy of self with changes applied.

        Parameters
        ----------
        **changes
            Keyword changes in the new copy.

        Returns
        -------
        <same type as self>
            A copy of self with ``changes`` incorporated.
        """
        return attrs.evolve(self, **changes)


# ------------------------------------------------------------------------------
#                         Kernel dictionary contexts
# ------------------------------------------------------------------------------


@define
class KernelContext(Context):
    """
    Kernel evaluation context data structure. This class is used *e.g.* to store
    information about the spectral configuration to apply when generating kernel
    dictionaries or update parameter maps associated with a
    :class:`.SceneElement` instance.
    """

    si: SpectralIndex = documented(
        attrs.field(
            factory=SpectralIndex.new,
            converter=SpectralIndex.convert,
            validator=attrs.validators.instance_of(SpectralIndex),
        ),
        doc="Spectral index (used to evaluate quantities with any degree "
        "or kind of dependency vs spectrally varying quantities).",
        type=":class:`.SpectralIndex`",
        init_type=":class:`.SpectralIndex` or dict",
        default=":meth:`SpectralIndex.new() <.SpectralIndex.new>`",
    )

    kwargs: dict[str, t.Any] = documented(
        attrs.field(factory=dict),
        doc="Object-specific parameter overrides.",
        type="dict",
        default="{}",
    )

    @property
    def index_formatted(self) -> str:
        return self.si.formatted_repr



from __future__ import annotations

import datetime
import functools
import inspect
import typing as t
import warnings

from packaging import version

from .numpydoc import format_doc, parse_doc
from .._version import _version

# Note: Most of this code is taken from https://github.com/briancurtin/deprecation.
# Class wrapping code is inspired by https://github.com/tantale/deprecated.


def deprecated(
    deprecated_in: str | None = None,
    removed_in: str | None = None,
    details: str = "",
    current_version: str = None,
):
    r"""
    Decorate a component to signify its deprecation

    This component wraps a component (function or class) that will soon be
    removed and does two things:

    * The docstring of the method will be modified to include a notice
      about deprecation, *e.g.* "Deprecated since 0.9.11. Use foo instead."
    * Raises a :class:`.DeprecatedWarning` via the :mod:`warnings` module,
      which is a subclass of the built-in :class:`DeprecationWarning`. Note that
      built-in :class:`DeprecationWarning`\ s are ignored by default, so for users
      to be informed of said warnings they will need to enable them---see
      the :mod:`warnings` module documentation for more details.

    Parameters
    ----------
    deprecated_in : str, optional, default: None
        The version at which the decorated method is considered deprecated.
        This will usually be the next version to be released when the decorator
        is added. The default is ``None``, which effectively means immediate
        deprecation. If this is not specified, then the `removed_in` and
        `current_version` arguments are ignored.

    removed_in : str, optional, default: None
        The version or :class:`datetime.date` when the decorated method will be
        removed. The default is ``None``, specifying that the component is not
        currently planned to be removed.
        Note: This parameter cannot be set to a value if `deprecated_in` is set
        to ``None``.

    details : str, optional
        Extra details to be added to the method docstring and warning. For
        example, the details may point users to a replacement method, such as
        "Use the foo_bar method instead". By default, there are no details.

    current_version : str, optional
        Current version. If unset, Eradiate's package version number is looked
        up automatically.

    Raises
    ------
    ValueError
        If `removed_in` is not ``None`` and `deprecated_in` is not ``None``.

    Examples
    --------
    This decorator allows for specifying the version from which the wrapped
    component is deprecated, and when it will be retired (this latter argument
    is optional):

    .. testsetup:: deprecated

       from eradiate.util.deprecation import deprecated

    .. doctest:: deprecated

       >>> @deprecated(deprecated_in="0.21.1", removed_in="0.22.1")
       ... def foo(): ...
       >>> foo()

    This will emit a :class:`.UnsupportedWarning`, which you can then handle.
    We can also deprecate classes:

    .. doctest:: deprecated

       >>> @deprecated(deprecated_in="0.21.1", removed_in="0.22.1")
       ... class SomeClass: ...

    If those components are used before v0.21.1, nothing happens. If they are
    used between v0.21.1 (included) and v0.22.1 (excluded), a
    :class:`.DeprecatedWarning` is emitted. If they are used from v0.22.1 on,
    an :class:`.UnsupportedWarning` is emitted.

    When used with static or class methods, their respective decorators must be
    applied *after* this one:

    .. doctest:: deprecated

       >>> class SomeClass:
       ...     @staticmethod
       ...     @deprecated(deprecated_in="0.21.1", removed_in="0.22.1")
       ...     def staticmethod(): ...
       ...
       ...     @classmethod
       ...     @deprecated(deprecated_in="0.21.1", removed_in="0.22.1")
       ...     def classmethod(cls): ...

    .. testcleanup:: deprecated

       del deprecated
    """
    # You can't just jump to removal. It's weird, unfair, and also makes
    # building up the docstring weird.
    if deprecated_in is None and removed_in is not None:
        raise ValueError(
            "Cannot set removed_in to a value without also setting deprecated_in"
        )

    # Only warn when it's appropriate. There may be cases when it makes sense
    # to add this decorator before a formal deprecation period begins.
    # In CPython, PendingDeprecatedWarning gets used in that period,
    # so perhaps mimic that at some point.
    is_deprecated = False
    is_unsupported = False

    # StrictVersion won't take a None or a "", so make whatever goes to it
    # is at least *something*. Compare versions only if removed_in is not
    # of type datetime.date
    if isinstance(removed_in, datetime.date):
        if datetime.date.today() >= removed_in:
            is_unsupported = True
        else:
            is_deprecated = True
    else:
        current_version = version.parse(
            _version if current_version is None else current_version
        )

        if removed_in is not None and current_version >= version.parse(removed_in):
            is_unsupported = True
        elif deprecated_in and current_version >= version.parse(deprecated_in):
            is_deprecated = True

    should_warn = any([is_deprecated, is_unsupported])

    def _wrapper(wrapped):
        if should_warn:
            # The various parts of this decorator being optional makes for
            # a number of ways the deprecation notice could go. The following
            # makes for a nicely constructed sentence with or without any
            # of the parts.

            # If removed_in is a date, use "removed on"
            # If removed_in is a version, use "removed in"
            parts = [".. deprecated::"]
            if deprecated_in:
                parts.append(f" {deprecated_in}")
            if removed_in:
                parts.append(
                    "\n   This will be removed "
                    f"{'on' if isinstance(removed_in, datetime.date) else 'in'} "
                    f"{removed_in}."
                )
            if details:
                parts.append(f"\n   {details}")

            deprecation_note = "".join(parts)

            # Parse and update wrapped object docstring
            doc = wrapped.__doc__ or ""  # Safety in case wrapped has no docstring
            sections = parse_doc(doc)
            sections["_deprecation"] = deprecation_note
            wrapped.__doc__ = "".join(format_doc(sections))

            # Define the actual emitted warning
            warning_cls = UnsupportedWarning if is_unsupported else DeprecatedWarning
            the_warning = warning_cls(
                wrapped.__name__, deprecated_in, removed_in, details
            )

        if inspect.isclass(
            wrapped
        ):  # We're wrapping a class: issue a warning upon class instantiation
            old_new = wrapped.__new__

            def wrapped_cls(cls, *args, **kwargs):
                if should_warn:
                    warnings.warn(
                        the_warning, category=DeprecationWarning, stacklevel=2
                    )

                if old_new is object.__new__:
                    return old_new(cls)

                else:  # actually, we don't know the real signature of *old_new*
                    return old_new(cls, *args, **kwargs)

            wrapped.__new__ = staticmethod(wrapped_cls)

            return wrapped

        else:  # We assume we're wrapping a function: issue a warning upon call

            @functools.wraps(wrapped)
            def _inner(*args, **kwargs):
                if should_warn:
                    warnings.warn(
                        the_warning, category=DeprecationWarning, stacklevel=2
                    )

                return wrapped(*args, **kwargs)

            return _inner

    return _wrapper


def substitute(subs: dict[str, tuple[type, dict[str, str]]]) -> t.Callable:
    """
    Generate a simple module :func:`__getattr__` which redirects outdated
    attribute lookups to current values with a deprecation warning.

    Parameters
    ----------
    subs: dict
        A dictionary with outdated names as keys. Values are 2-tuples consisting
        of the current substitute and a dictionary of keyword arguments passed
        to the :class:`DeprecatedWarning` constructor.

    Returns
    -------
    callable
        The generated module :func:`__getattr__` function.

    Example
    -------
    >>> __getattr__ = substitute(
    ...     {
    ...         "OneDimExperiment": (  # Old name
    ...             AtmosphereExperiment,  # New type
    ...             {"deprecated_in": "0.22.5", "removed_in": "0.22.7"},  # Keyword args
    ...         )
    ...     }
    ... )

    """

    def __getattr__(name):
        if name in subs:
            new_type, kwargs = subs[name]
            warnings.warn(
                DeprecatedWarning(
                    component=f"'{name}'",
                    details=f"Use '{new_type.__name__}' instead.",
                    **kwargs,
                ),
            )
            return new_type

        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

    return __getattr__


class DeprecatedWarning(DeprecationWarning):
    """
    A warning class for deprecated methods

    This is a specialization of the built-in :class:`DeprecationWarning`,
    adding parameters that allow us to get information into the ``__str__``
    that ends up being sent through the :mod:`warnings` system.
    The attributes aren't able to be retrieved after the warning gets
    raised and passed through the system as only the class--not the
    instance--and message are what gets preserved.

    Parameters
    ----------
    component : type or callable
        The component being deprecated.

    deprecated_in : str, optional, default: None
        The version that `component` is deprecated in.

    removed_in : str or datetime, optional, default: None
        The version or :class:`datetime.date` specifying when `component` gets
        removed.

    details : str, optional
        Optional details about the deprecation. Most often this will include
        directions on what to use instead of the now deprecated code.
    """

    def __init__(self, component, deprecated_in, removed_in, details=""):
        # NOTE: The docstring only works for this class if it appears up
        # near the class name, not here inside __init__. I think it has
        # to do with being an exception class.
        self.component = component
        self.deprecated_in = deprecated_in
        self.removed_in = removed_in
        self.details = details
        super(DeprecatedWarning, self).__init__(
            component, deprecated_in, removed_in, details
        )

    def __str__(self):
        parts = [f"{self.component} is deprecated"]
        if self.deprecated_in:
            parts.append(f" as of {self.deprecated_in}")
        if self.removed_in:
            parts.append(
                " and will be removed "
                f"{'on' if isinstance(self.removed_in, datetime.date) else 'in'} "
                f"{self.removed_in}"
            )
        if any([self.deprecated_in, self.removed_in, self.details]):
            parts.append(".")
        if self.details:
            parts.append(f" {self.details}")

        return "".join(parts)


class UnsupportedWarning(DeprecatedWarning):
    """
    A warning class for methods to be removed

    This is a subclass of :class:`.DeprecatedWarning` and is used to output a
    proper message about a component being unsupported.
    """

    def __str__(self):
        parts = [f"{self.component} is unsupported as of {self.removed_in}."]
        if self.details:
            parts.append(f" {self.details}")

        return "".join(parts)



from ._core import AbstractHeterogeneousAtmosphere as AbstractHeterogeneousAtmosphere
from ._core import Atmosphere as Atmosphere
from ._core import atmosphere_factory as atmosphere_factory
from ._heterogeneous import HeterogeneousAtmosphere as HeterogeneousAtmosphere
from ._homogeneous import HomogeneousAtmosphere as HomogeneousAtmosphere
from ._molecular import MolecularAtmosphere as MolecularAtmosphere
from ._particle_dist import ArrayParticleDistribution as ArrayParticleDistribution
from ._particle_dist import (
    ExponentialParticleDistribution as ExponentialParticleDistribution,
)
from ._particle_dist import GaussianParticleDistribution as GaussianParticleDistribution
from ._particle_dist import (
    InterpolatorParticleDistribution as InterpolatorParticleDistribution,
)
from ._particle_dist import ParticleDistribution as ParticleDistribution
from ._particle_dist import UniformParticleDistribution as UniformParticleDistribution
from ._particle_dist import (
    particle_distribution_factory as particle_distribution_factory,
)
from ._particle_layer import ParticleLayer as ParticleLayer
from ._util import eval_transmittance_ckd as eval_transmittance_ckd
from ._util import eval_transmittance_mono as eval_transmittance_mono



"""*attrs*-based utility classes and functions."""

from __future__ import annotations

import enum
import re
import typing as t
from textwrap import dedent, indent

import attrs

from .util import numpydoc


class _Auto:
    """
    Sentinel class to indicate when a dynamic field value is expected to be
    set automatically. ``_Auto`` is a singleton. There is only ever one of it.

    Notes
    -----
    ``bool(_Auto)`` evaluates to ``False``.
    """

    _singleton = None

    def __new__(cls):
        if _Auto._singleton is None:
            _Auto._singleton = super(_Auto, cls).__new__(cls)
        return _Auto._singleton

    def __repr__(self):
        return "AUTO"

    def __bool__(self):
        return False

    def __len__(self):
        return 0


#: Typing alias for :data:`.AUTO`.
AutoType = _Auto

#: Sentinel to indicate when a dynamic field value is expected to be set
#: automatically.
AUTO = _Auto()


class MetadataKey(enum.Enum):
    """
    Attribute metadata keys.

    These Enum values should be used as metadata attribute keys.
    """

    DOC = "doc"  #: Documentation for this field (str)
    TYPE = "type"  #: Documented type for this field (str)
    INIT_TYPE = "init_type"  #: Documented constructor parameter for this field (str)
    DEFAULT = "default"  #: Documented default value for this field (str)


class DocFlags(enum.Flag):
    """
    Extra flags used to pass information about docs.
    """

    NOINIT = 0


# ------------------------------------------------------------------------------
#                           Attribute docs extension
# ------------------------------------------------------------------------------


@attrs.define
class _FieldDoc:
    """Internal convenience class to store field documentation information."""

    doc = attrs.field(default=None)
    type = attrs.field(default=None)
    init_type = attrs.field(default=None)
    default = attrs.field(default=None)


def _eradiate_formatter(cls_doc: str, field_docs: dict[str, _FieldDoc]) -> str:
    """
    Appends a section on attributes to a class docstring.
    This docstring formatter is appropriate for Eradiate's current docstring
    format.

    Parameters
    ----------
    cls_doc : str
        Class docstring to extend.

    field_docs : dict[str, _FieldDoc]
        Attributes documentation content.

    Returns
    -------
    str
        Updated class docstring.
    """
    # Do nothing if field is not documented
    if not field_docs:
        return cls_doc

    docstrings = []

    # Create docstring entry for each documented field
    for field_name, field_doc in field_docs.items():
        type_doc = f": {field_doc.type}" if field_doc.type is not None else ""
        default_doc = f" = {field_doc.default}" if field_doc.default is not None else ""

        docstrings.append(
            f"``{field_name.lstrip('_')}``{type_doc}{default_doc}\n"
            f"{indent(field_doc.doc, '    ')}\n"
        )

    # Assemble entries
    if docstrings:
        if cls_doc is None:
            cls_doc = ""

        return "\n".join(
            (
                dedent(cls_doc.lstrip("\n")).rstrip(),
                "",
                ".. rubric:: Constructor arguments / instance attributes",
                "",
                "\n".join(docstrings),
                "",
            )
        )
    else:
        return cls_doc


def _numpy_formatter(cls_doc: str, field_docs: dict[str, _FieldDoc]) -> str:
    """
    Append a section on attributes to a class docstring.
    This docstring formatter is appropriate for the Numpy docstring format.

    Parameters
    ----------
    cls_doc : str
        Class docstring to extend.

    field_docs : dict[str, _FieldDoc]
        Attributes documentation content.

    Returns
    -------
    str
        Updated class docstring.
    """
    # Do nothing if no field is documented
    if not field_docs:
        return cls_doc

    param_docstrings = []
    attr_docstrings = []

    # Create docstring entry for each documented field
    for field_name, field_doc in field_docs.items():
        field_type = field_doc.type
        field_init_type = field_doc.init_type

        # Generate constructor parameter docstring entry
        if field_init_type is not DocFlags.NOINIT:
            init_type_doc = (
                f" : {field_init_type}" if field_init_type is not None else ""
            )
            default_doc = (
                f", default: {field_doc.default}"
                if field_doc.default is not None
                else ""
            )
            param_docstrings.append(
                f"{field_name.lstrip('_')}{init_type_doc}{default_doc}\n"
                f"{indent(field_doc.doc, '    ')}\n"
            )

            # Generate attribute docstring entry
            type_doc = field_type if field_type is not None else ""
            if not field_name.startswith("_"):
                field_doc_brief = re.split(r"\. |\.\n", field_doc.doc)[0].strip()
                if not field_doc_brief.endswith("."):
                    field_doc_brief += "."
                attr_docstrings.append(
                    f"{field_name} : {type_doc}\n{indent(field_doc_brief, '    ')}\n"
                )

    # Assemble entries
    if param_docstrings or attr_docstrings:
        if cls_doc is None:
            cls_doc = ""

        cls_doc = dedent(cls_doc.lstrip("\n")).rstrip()

        # Parse docstrings
        sections = numpydoc.parse_doc(cls_doc)

        # Append generated docstrings to the relevant section
        sections["Parameters"] = "\n".join(param_docstrings) + sections.get(
            "Parameters", ""
        )
        sections["Fields"] = "\n".join(attr_docstrings) + sections.get("Fields", "")

        # Generate section full text
        doc = numpydoc.format_doc(sections)
        return doc

    else:
        return cls_doc


def parse_docs(cls: type) -> type:
    """
    Extract attribute documentation and update class docstring with it.

    This decorator will examine each *attrs* attribute and check its metadata
    for documentation content. It will then update the class's docstring
    based on this content.

    Parameters
    ----------
    cls : type
        Class whose attributes should be processed.

    Returns
    -------
    type
        Updated class.

    See Also
    --------
    :func:`documented` : Field documentation definition function.

    Notes
    -----
    * Meant to be used as a class decorator.
    * Must be applied **after** :func:`@attr.s <attr.s>` (or any other
      attrs decorator).
    * Fields must be documented using :func:`documented`.
    """
    formatter = _numpy_formatter

    docs = {}
    for field in cls.__attrs_attrs__:
        # Collect documentation entries
        if MetadataKey.DOC in field.metadata:
            # Collect field docstring
            docs[field.name] = _FieldDoc(doc=field.metadata[MetadataKey.DOC])

            # Collect field type
            if MetadataKey.TYPE in field.metadata:
                docs[field.name].type = field.metadata[MetadataKey.TYPE]
            else:
                docs[field.name].type = str(field.type)

            # Collect field init type
            if MetadataKey.INIT_TYPE in field.metadata:
                docs[field.name].init_type = field.metadata[MetadataKey.INIT_TYPE]
            elif field.init is False:
                docs[field.name].init_type = DocFlags.NOINIT
            else:
                docs[field.name].init_type = docs[field.name].type

            # Collect default value
            if MetadataKey.DEFAULT in field.metadata:
                docs[field.name].default = field.metadata[MetadataKey.DEFAULT]

    # Update docstring
    cls.__doc__ = formatter(cls.__doc__, docs)

    return cls


def documented(
    attrib: attrs.Attribute,
    doc: str | None = None,
    type: str | None = None,
    init_type: str | None = None,
    default: str | None = None,
) -> attrs.Attribute:
    """
    Declare an attrs field as documented.

    Parameters
    ----------
    attrib : attrs.Attribute
        *attrs* attribute definition to which documentation is to be attached.

    doc : str, optional
        Docstring for the considered field. If set to ``None``, this function
        does nothing.

    type : str, optional
        Documented type for the considered field.

    init_type : str, optional
        Documented constructor parameter for the considered field.

    default : str, optional
        Documented default value for the considered field.

    Returns
    -------
    attrs.Attribute
        ``attrib``, with metadata updated with documentation contents.

    See Also
    --------
    :func:`attrs.field`, :func:`pinttr.field`, :func:`parse_docs`
    """
    if doc is not None:
        attrib.metadata[MetadataKey.DOC] = doc

    if type is not None:
        attrib.metadata[MetadataKey.TYPE] = type

    if init_type is not None:
        attrib.metadata[MetadataKey.INIT_TYPE] = init_type

    if default is not None:
        attrib.metadata[MetadataKey.DEFAULT] = default

    return attrib


def get_doc(
    cls: type, attrib: str, field: t.Literal["doc", "type", "init_type", "default"]
) -> str:
    """
    Fetch attribute documentation field. Requires field metadata to be processed
    with :func:`documented`.

    Parameters
    ----------
    cls : type
        Class from which to get the attribute.

    attrib : str
        Attribute from which to get the doc field.

    field : {"doc", "type", "init_type", "default"}
        Documentation field to query.

    Returns
    -------
    str
        Queried documentation content.

    Raises
    ------
    ValueError
        If the requested ``field`` is missing from the target attribute's
        metadata.

    ValueError
        If the requested ``field`` is unsupported.
    """
    try:
        key = MetadataKey[field.upper()]
    except ValueError:
        raise ValueError(f"unsupported attribute doc field {field}")

    try:
        return attrs.fields_dict(cls)[attrib].metadata[key]
    except KeyError:
        raise ValueError(f"{cls.__name__}.{attrib} has no documented field '{field}'")

    # This is very unlikely to happen, but just in case, we raise
    raise RuntimeError


def define(maybe_cls=None, **kwargs):
    """
    A wrapper around :func:`attrs.define` that automatically applies docstring
    processing with :func:`.parse_docs`. All arguments are forwarded to
    :func:`attrs.define`.
    """

    def wrap(cls):
        return parse_docs(attrs.define(maybe_cls=cls, **kwargs))

    return wrap if maybe_cls is None else wrap(maybe_cls)


def frozen(maybe_cls=None, **kwargs):
    """
    A wrapper around :func:`attrs.frozen` that automatically applies docstring
    processing with :func:`.parse_docs`. All arguments are forwarded to
    :func:`attrs.frozen`.
    """

    def wrap(cls):
        return parse_docs(attrs.frozen(maybe_cls=cls, **kwargs))

    return wrap if maybe_cls is None else wrap(maybe_cls)



"""
Data loading, conversion and output components.
"""

from __future__ import annotations

import warnings
from typing import Literal

import numpy as np
import pint
import pinttrs
import xarray as xr

from ..data import fresolver
from ..typing import PathLike
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg


def _convert_units(value: str):
    if value == "per cent":
        return "percent"
    return value


def _get_units(ds, var, fallback_units=None):
    if "units" in ds[var].attrs:
        return ureg.Unit(_convert_units(ds[var].attrs["units"]))
    elif fallback_units is not None and var in fallback_units:
        return ureg.Unit(_convert_units(fallback_units[var]))
    else:
        raise ValueError(
            "load_aerosol_libradtran(): The input dataset specifies no units "
            f"for variable '{var}'; this can be addressed by passing them through "
            "the 'fallback_units' parameter."
        )


def load_aerosol_libradtran(
    data: PathLike | xr.Dataset,
    particle_shape: Literal["spherical", "spheroidal"] | None = None,
    tolerance: dict[str, pint.Quantity | float] | None = None,
    wbounds: tuple = (None, None),
    fallback_units: dict[str, str] | None = None,
    **kwargs,
) -> xr.Dataset:
    """
    Convert a libRadtran NetCDF aerosol file to the Eradiate aerosol data format.
    Both effective radius and humidity-indexed data are supported.

    Parameters
    ----------
    data : Dataset or path-like
        A libRadtran NetCDF aerosol dataset. If a path is passed, it will be
        resolved by the file resolver and tentatively loaded into memory.

    particle_shape : {"spherical", "spheroidal"}, optional
        The expected shape of the particle (this will tell the phase matrix
        coefficients it should expect). If unset, the shape is inferred from the
        input dataset.

    reff, hum : float or quantity
        For datasets with a humidity or effective radius dimension, the
        coordinate point to select. By default, the nearest data point is
        selected; tolerance can be adjusted using the ``tolerance`` parameter.
        Quantities are accepted (see Notes for expected dimensions and
        defaults). These parameters are optional if the dataset has only one
        point on these dimensions.

    wbounds : tuple, optional
        Bounds to restrict the spectral domain where conversion is performed.
        This parameter accepts a tuple with the minimum and maximum values
        (use ``None`` to leave a bound open). Examples:
        ``(300, 3000)``, ``(None, 3000)``, ``(0.3, 3) * ureg.micron``,
        ``(None, 3 * ureg.micron)``.

    tolerance : dict
        A mapping that allows to specify a tolerance for nearest neighbour
        lookup for relevant parameters. Units are applied with the same rules as
        for the ``reff`` and ``hum`` parameters.

    fallback_units : dict, optional
        A mapping that specifies units to apply to variables that are missing them.

    Returns
    -------
    Dataset

    Notes
    -----
    .. list-table::
        :widths: 1 2 2
        :header-rows: 1

        * - Parameter
          - Dimension
          - Default units
        * - ``reff``
          - Length
          - μm
        * - ``hum``
          - Dimensionless (fraction)
          - percent
        * - ``wbounds``
          - Wavelength
          - ``ucc.get("wavelength")`` (usually, nm)

    * The current implementation resamples the angular dimension at the
      highest resolution to minimize the loss of information on phase matrix
      coefficients.

    * All conversion is done in memory: very large dataset might result in
      massive converted data. In such case, an easy way to split the conversion
      is to chunk it on the spectral dimension.
    """
    VARS_TO_DIMS = {"wavelen": "nlam", "reff": "nreff", "hum": "nhum"}
    KWARG_TO_DEFAULT_UNITS = {
        "w": ucc.get("wavelength"),
        "hum": ureg.Unit("percent"),
        "reff": ureg.Unit("micrometer"),
    }

    # Load aerosol component dataset
    if not isinstance(data, xr.Dataset):
        data = fresolver.load_dataset(data)

    # Extract required variables
    vars = ["phase", "ext", "ssa", "theta", "wavelen"]
    for var in ["reff", "hum"]:
        if var in data:
            vars.append(var)
    data = data[vars]

    # Select on humidity and effective radius
    if tolerance is None:
        tolerance = {}

    for kwarg in ["hum", "reff"]:
        var = kwarg
        if var not in data:
            continue

        units = _get_units(data, var, fallback_units)
        da = data[var]
        default_units = KWARG_TO_DEFAULT_UNITS[kwarg]

        if len(da) > 1:
            if kwarg not in kwargs:
                raise TypeError(
                    f"load_aerosol_libradtran() is missing keyword argument '{kwarg}' "
                    f"(allowed: {da.values})"
                )
            else:
                kwarg_value = pinttrs.converters.ensure_units(
                    np.atleast_1d(kwargs.pop(kwarg)), default_units=default_units
                ).to(units)
        else:
            if kwarg not in kwargs:
                kwarg_value = da.values * units
            else:
                kwarg_value = pinttrs.converters.ensure_units(
                    np.atleast_1d(kwargs.pop(kwarg)), default_units=default_units
                ).to(units)

        sel_kwargs = {var: kwarg_value.m, "method": "nearest"}
        if kwarg in tolerance:
            sel_kwargs["tolerance"] = tolerance[kwarg]

        data = data.swap_dims({VARS_TO_DIMS[var]: var})
        data = data.sel(**sel_kwargs)
        data = data.squeeze(var, drop=True)

    if kwargs:
        warnings.warn(
            "load_aerosol_libradtran() got unexpected keyword arguments "
            f"{list(kwargs.keys())}, which were not used"
        )

    # Filter wavelengths if requested
    wmin, wmax = wbounds

    units = _get_units(data, "wavelen", fallback_units)
    default_units = KWARG_TO_DEFAULT_UNITS["w"]

    if wmin is not None:
        wmin = pinttrs.converters.ensure_units(wmin, default_units=default_units).m_as(
            units
        )
        data = data.where(data["wavelen"] >= wmin).dropna("nlam", how="all")
    if wmax is not None:
        wmax = pinttrs.converters.ensure_units(wmax, default_units=default_units).m_as(
            units
        )
        data = data.where(data["wavelen"] <= wmax).dropna("nlam", how="all")

    wavelength = data["wavelen"].values * units

    # Phase function
    if particle_shape is None:
        if len(data["nphamat"] == 4):
            particle_shape = "spherical"
        elif len(data["nphamat"] == 6):
            particle_shape = "spheroidal"
        else:
            raise ValueError("Could not detect particle shape type")

    if particle_shape == "spherical":
        ij_to_nphamat = {
            (0, 0): 0,
            (1, 1): 0,
            (0, 1): 1,
            (1, 0): 1,
            (2, 2): 2,
            (3, 3): 2,
            (2, 3): 3,
            (3, 2): 3,
        }
    elif particle_shape == "spheroidal":
        ij_to_nphamat = {
            (0, 0): 0,
            (0, 1): 1,
            (1, 0): 1,
            (1, 1): 4,
            (2, 2): 2,
            (2, 3): 3,
            (3, 2): 3,
            (3, 3): 5,
        }
    else:
        raise NotImplementedError(f"Unknown particle shape '{particle_shape}'")

    # -- Create angular grid (highest resolution possible)
    mus = np.cos(np.deg2rad(data["theta"].values.ravel()))
    mus = mus[~np.isnan(mus)]
    mus = np.unique(mus)

    # -- Resample all phase matrix components and fill data array with shape
    #    [wavelength, theta, i, j]
    n_wavelength = len(wavelength)
    n_theta = len(mus)
    phase_np = np.zeros((n_wavelength, n_theta, 4, 4))

    for i_wavelength in range(n_wavelength):
        for (i, j), nphamat in ij_to_nphamat.items():
            data_selected = data.isel(nlam=i_wavelength, nphamat=nphamat).dropna(
                "nthetamax"
            )
            x = mus
            xp = np.cos(np.deg2rad(data_selected["theta"].values)).ravel()
            fp = data_selected["phase"].values
            p = np.interp(x, xp, fp)
            phase_np[i_wavelength, :, i, j] = p

    # Populate Eradiate dataset with correct format
    phase_eradiate = xr.Dataset(
        data_vars={
            "sigma_t": (["w"], data["ext"].values, {"units": "1/km"}),
            "albedo": (["w"], data["ssa"].values, {"units": ""}),
            "phase": (["w", "mu", "i", "j"], phase_np),
        },
        coords={
            "w": ("w", wavelength.m_as("nm"), {"units": "nm"}),
            "mu": ("mu", mus),
            "i": ("i", range(4)),
            "j": ("j", range(4)),
        },
    )

    return phase_eradiate



from . import io as io
from ._asset_manager import AssetManager as AssetManager
from ._asset_manager import asset_manager as asset_manager
from ._file_resolver import FileResolver as FileResolver
from ._file_resolver import fresolver as fresolver



from __future__ import annotations

from abc import ABC

import attrs

from ..core import NodeSceneElement
from ..._factory import Factory
from ...attrs import define, documented, get_doc

integrator_factory = Factory()
integrator_factory.register_lazy_batch(
    [
        ("_path_tracers.PathIntegrator", "path", {}),
        ("_path_tracers.VolPathIntegrator", "volpath", {}),
        ("_path_tracers.VolPathMISIntegrator", "volpathmis", {}),
        ("_path_tracers.PiecewiseVolPathIntegrator", "piecewise_volpath", {}),
    ],
    cls_prefix="eradiate.scenes.integrators",
)


@define(eq=False, slots=False)
class Integrator(NodeSceneElement, ABC):
    """
    Abstract base class for all integrator elements.

    Notes
    -----
    * This class is to be used as a mixin.
    """

    id: str | None = documented(
        attrs.field(
            default="integrator",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(NodeSceneElement, "id", "doc"),
        type=get_doc(NodeSceneElement, "id", "type"),
        init_type=get_doc(NodeSceneElement, "id", "init_type"),
        default='"integrator"',
    )

    timeout: float | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(float),
        ),
        doc="Maximum amount of time to spend during a kernel render pass in "
        "second (excluding scene parsing).",
        type="float or None",
        init_type="float, optional",
        default="None",
    )

    moment: bool = documented(
        attrs.field(
            default=False,
            converter=attrs.converters.optional(bool),
        ),
        doc="If ``True``, calculates the variance of the output film.",
        type="bool",
        init_type="bool, optional",
        default="False",
    )

    stokes: bool = documented(
        attrs.field(
            default=False,
            converter=attrs.converters.optional(bool),
        ),
        doc="If ``True``, calculate and output the polarized components of "
        "the Stokes vector. Note that this will only work with a polarized "
        "mode and a sensor that measures radiance.",
        type="bool",
        init_type="bool, optional",
        default="False",
    )

    meridian_align: bool = documented(
        attrs.field(
            default=True,
            converter=attrs.converters.optional(bool),
        ),
        doc="If ``True``, align the Stokes vector to the meridian plane. "
        "Otherwise, default to aligning to the sensor's x-axis. This parameter "
        "is only relevant in polarized mode, when measuring radiance with "
        "``stokes=True``.",
        type="bool",
        init_type="bool, optional",
        default="True",
    )



from __future__ import annotations

import datetime
import logging
import typing as t
import warnings
from abc import ABC, abstractmethod
from collections.abc import Sequence

import attrs
import mitsuba as mi
import numpy as np
import pinttr
import xarray as xr
from hamilton.driver import Driver

import eradiate

from .. import converters, validators
from .. import pipelines as pl
from ..attrs import AUTO, define, documented, frozen
from ..contexts import KernelContext
from ..exceptions import UnsupportedModeError
from ..kernel import (
    KernelDict,
    KernelSceneParameterMap,
    MitsubaObjectWrapper,
    mi_render,
    mi_traverse,
)
from ..quad import Quad
from ..rng import SeedState
from ..scenes.core import Scene, SceneElement, get_factory, traverse
from ..scenes.illumination import (
    AbstractDirectionalIllumination,
    ConstantIllumination,
    DirectionalIllumination,
    illumination_factory,
)
from ..scenes.integrators import Integrator, integrator_factory
from ..scenes.measure import (
    Measure,
    MultiDistantMeasure,
    measure_factory,
)
from ..spectral.ckd_quad import CKDQuadConfig
from ..spectral.grid import CKDSpectralGrid, MonoSpectralGrid, SpectralGrid
from ..spectral.index import CKDSpectralIndex, MonoSpectralIndex, SpectralIndex
from ..units import unit_registry as ureg
from ..util.misc import deduplicate_sorted

logger = logging.getLogger(__name__)


@frozen(init=False)
class MeasureRegistry(Sequence):
    """
    A simple list wrapper holding measures, with additional lookup methods and
    metadata.

    The constructor converts dictionaries automatically and checks for duplicate
    IDs (raises a :class:`ValueError` if it finds any).
    """

    _measures: list[Measure] = attrs.field(
        converter=list,
        validator=attrs.validators.deep_iterable(
            member_validator=attrs.validators.instance_of(Measure)
        ),
    )
    _id_to_idx: dict[str, int] = attrs.field(factory=dict, repr=False)
    _idx_to_id: dict[int, str] = attrs.field(factory=dict, repr=False)

    def __init__(self, measures: t.Sequence):
        # Convert all values to a measure
        measures = [measure_factory.convert(x) for x in measures]

        # Check for duplicate IDs
        values = dict(zip(*np.unique([m.id for m in measures], return_counts=True)))
        duplicate_ids = [k for k, v in values.items() if v > 1]
        if duplicate_ids:
            raise ValueError(f"duplicate measure ids: {duplicate_ids}")

        # Update index and ID registries
        id_to_idx = {m.id: i for i, m in enumerate(measures)}
        idx_to_id = {i: m.id for i, m in enumerate(measures)}

        # Finalize initialization
        self.__attrs_init__(measures=measures, id_to_idx=id_to_idx, idx_to_id=idx_to_id)

    def __getitem__(self, index):
        return self._measures[index]

    def __len__(self):
        return len(self._measures)

    def get_index(self, value: str | int) -> int:
        """
        Get the index corresponding to a given measure ID. Integers are passed
        through.
        """
        if isinstance(value, str):
            return self._id_to_idx[value]
        elif isinstance(value, int):
            return value
        else:
            raise TypeError(f"unhandled type {type(value)}")

    def get_id(self, value: str | int) -> str:
        """
        Get the ID corresponding to a given index. Strings are passed through.
        """
        if isinstance(value, str):
            return value
        elif isinstance(value, int):
            return self._idx_to_id[value]
        else:
            raise TypeError(f"unhandled type {type(value)}")

    def resolve(self, value: str | int) -> Measure:
        """
        Resolve a measure based on its ID or index.
        """
        return self[self.get_index(value)]


@define
class Experiment(ABC):
    """
    Abstract base class for all Eradiate experiments. An experiment consists of
    a high-level scene specification parametrized by natural user input, a
    processing and post-processing pipeline, and a result storage data
    structure.
    """

    # Internal Mitsuba scene. This member is not set by the end-user, but rather
    # by the Experiment itself during initialization.
    mi_scene: MitsubaObjectWrapper | None = attrs.field(
        default=None,
        repr=False,
        init=False,
    )

    measures: MeasureRegistry = documented(
        attrs.field(
            factory=lambda: MeasureRegistry([MultiDistantMeasure()]),
            converter=lambda value: MeasureRegistry(
                pinttr.util.always_iterable(value)
                if not isinstance(value, dict)
                else [measure_factory.convert(value)]
            ),
        ),
        doc="List of measure specifications. The passed list may contain "
        "dictionaries, which will be interpreted by "
        ":data:`.measure_factory`. "
        "Optionally, a single :class:`.Measure` or dictionary specification "
        "may be passed and will automatically be wrapped into a list.",
        type=".MeasureRegistry",
        init_type="list of :class:`.Measure` or list of dict or "
        ":class:`.Measure` or dict",
        default=":class:`MultiDistantMeasure() <.MultiDistantMeasure>`",
    )

    integrator: Integrator = documented(
        attrs.field(
            default=AUTO,
            converter=converters.auto_or(integrator_factory.convert),
            validator=validators.auto_or(
                attrs.validators.instance_of(Integrator),
            ),
        ),
        doc="Monte Carlo integration algorithm specification. "
        "This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.integrator_factory`."
        "The integrator defaults to :data:`AUTO`, which will choose the appropriate "
        "integrator depending on the experiment's configuration. ",
        type=":class:`.Integrator` or AUTO",
        init_type=":class:`.Integrator` or dict or AUTO",
        default="AUTO",
    )

    # Storage for results, for each computed measure
    _results: dict[str, xr.Dataset] = attrs.field(factory=dict, repr=False)

    @property
    def results(self) -> dict[str, xr.Dataset]:
        """
        Post-processed simulation results.

        Returns
        -------
        dict[str, Dataset]
            Dictionary mapping measure IDs to xarray datasets.
        """
        return self._results

    _background_spectral_grid: SpectralGrid = documented(
        attrs.field(
            default=AUTO,
            validator=validators.auto_or(attrs.validators.instance_of(SpectralGrid)),
            repr=False,
        ),
        doc="Background spectral grid. "
        "If the value is :data:`.AUTO`, the background spectral grid is "
        "automatically generated depending on the active mode and internal "
        "experiment constraints. Otherwise, the value must be convertible to "
        "a :class:`.SpectralGrid` instance.",
        type=".SpectralGrid or AUTO",
        init_type=".SpectralGrid or AUTO",
        default="AUTO",
    )

    @property
    def background_spectral_grid(self) -> SpectralGrid:
        return self._background_spectral_grid

    # Grid used to walk the spectral dimension for each measure.
    # Set upon initialization by the '_normalize_spectral()' method.
    _spectral_grids: dict[int, SpectralGrid] = attrs.field(
        factory=dict, init=False, repr=False
    )

    @property
    def spectral_grids(self) -> dict[int, SpectralGrid]:
        """
        A dictionary mapping measure index to the associated spectral grid.
        """
        return self._spectral_grids

    ckd_quad_config: CKDQuadConfig = documented(
        attrs.field(
            factory=lambda: CKDQuadConfig(ng_max=16),
            converter=CKDQuadConfig.convert,
            validator=attrs.validators.instance_of(CKDQuadConfig),
        ),
        doc="CKD quadrature rule generation configuration.",
        type=".CKDQuadConfig",
        init_type=".CKDQuadConfig or dict",
    )

    # CKD quadrature configuration for each bin.
    # Set upon initialization by the '_normalize_spectral()' method.
    _ckd_quads: dict[int, list[Quad]] = attrs.field(
        factory=dict, init=False, repr=False
    )

    @property
    def ckd_quads(self) -> dict[int, list[Quad]]:
        """
        A dictionary mapping measure index to the associated CKD quadrature rule
        (if relevant).
        """
        return self._ckd_quads

    def __attrs_post_init__(self):
        self._normalize_spectral()

    def _normalize_spectral(self) -> None:
        """
        Assemble a spectral grid based on the various elements in the scene.
        """
        # Collect atmosphere-based grid if relevant
        atmosphere = getattr(self, "atmosphere", None)
        abs_db = None
        if atmosphere is not None:
            abs_db = getattr(atmosphere, "absorption_data", None)
        if abs_db is not None:
            if self._background_spectral_grid is not AUTO:
                warnings.warn(
                    "User-specified a background spectral grid is overridden by "
                    "atmosphere spectral grid."
                )
            self._background_spectral_grid = SpectralGrid.from_absorption_database(
                atmosphere.absorption_data
            )

        # If needed, set the background grid
        if self._background_spectral_grid is AUTO:
            self._background_spectral_grid = SpectralGrid.default()

        # Select subparts of the grid that are covered by the SRF
        self._spectral_grids = {
            i: self.background_spectral_grid.select(measure.srf)
            for i, measure in enumerate(self.measures)
        }

        # Get quadrature rules for all bins
        ckd_quads = {}
        for i, measure in enumerate(self.measures):
            if eradiate.mode().is_ckd:
                spectral_grid: CKDSpectralGrid = self._spectral_grids[i]
                ckd_quads[i] = [
                    x[1] for x in spectral_grid.walk_quads(self.ckd_quad_config, abs_db)
                ]
            else:
                ckd_quads[i] = []
        self._ckd_quads = ckd_quads

    def clear(self) -> None:
        """
        Clear previous experiment results and reset internal state.
        """
        self.results.clear()

        for measure in self.measures:
            measure.mi_results.clear()

    @abstractmethod
    def init(self, drop_parameters: bool = True) -> None:
        """
        Generate kernel dictionary and initialize Mitsuba scene.

        Parameters
        ----------
        drop_parameters : bool
            If ``True``, drop Mitsuba scene parameters that are not used (*i.e.*
            that do not have an updater associated).
        """
        pass

    @abstractmethod
    def process(
        self,
        measures: None | int | list[int] = None,
        spp: int = 0,
        seed_state: SeedState | None = None,
    ) -> None:
        """
        Run simulation and collect raw results.

        Parameters
        ----------
        measures : int or list of int, optional
            Indices of the measures that will be processed. By default, all
            measures are processed.

        spp : int, optional
            Sample count. If set to 0, the value set in the original scene
            definition takes precedence.

        seed_state : :class:`.SeedState`, optional
            Seed state used to generate seeds to initialize Mitsuba's RNG at
            every iteration of the parametric loop. If unset, Eradiate's
            :attr:`root seed state <.root_seed_state>` is used.
        """
        pass

    @abstractmethod
    def postprocess(self, measures: None | int | list[int] = None) -> None:
        """
        Post-process raw results and store them in :attr:`results`.

        Parameters
        ----------
        measures : int or list of int, optional
            Indices of the measures that will be processed. By default, all
            measures are processed.
        """
        pass

    @abstractmethod
    def pipeline(self, measure: Measure | int) -> Driver:
        """
        Return the post-processing pipeline for a given measure.

        Parameters
        ----------
        measure : .Measure or int
            Measure for which the pipeline is generated.

        Returns
        -------
        hamilton.driver.Driver
        """
        pass

    @abstractmethod
    def context_init(self) -> KernelContext:
        """
        Return a single context used for scene initialization.

        Returns
        -------
        .KernelContext
        """
        pass

    @abstractmethod
    def contexts(self, measures: None | int | list[int] = None) -> list[KernelContext]:
        """
        Return a list of contexts used for processing.

        Parameters
        ----------
        measures : int or list of int, optional
            A list of the indexes of the measures to account for when emitting
            kernel contexts. If unset, all measures are accounted for.

        Returns
        -------
        list of .KernelContext
        """
        pass


def _extra_objects_converter(value: dict | None) -> dict:
    if not value:
        return {}

    result = {}

    for key, element_spec in value.items():
        if isinstance(element_spec, dict):
            element_spec = element_spec.copy()
            element_type = element_spec.pop("factory")
            factory = get_factory(element_type)
            result[key] = factory.convert(element_spec)

        else:
            result[key] = element_spec

    return result


@define
class EarthObservationExperiment(Experiment, ABC):
    """
    Abstract based class for experiments illuminated by a distant directional
    emitter.
    """

    extra_objects: dict[str, SceneElement] = documented(
        attrs.field(
            default=None,
            converter=_extra_objects_converter,
            validator=attrs.validators.deep_mapping(
                key_validator=attrs.validators.instance_of(str),
                value_validator=attrs.validators.instance_of(SceneElement),
            ),
        ),
        doc="Dictionary of extra objects to be added to the scene. "
        "The keys of this dictionary are used to identify the objects "
        "in the kernel dictionary.",
        type="dict",
        init_type="dict or None",
        default="None",
    )

    illumination: AbstractDirectionalIllumination | ConstantIllumination = documented(
        attrs.field(
            factory=DirectionalIllumination,
            converter=illumination_factory.convert,
            validator=attrs.validators.instance_of(
                (AbstractDirectionalIllumination, ConstantIllumination)
            ),
        ),
        doc="Illumination specification. "
        "This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.illumination_factory`.",
        type=":class:`.AbstractDirectionalIllumination` or "
        ":class:`.ConstantIllumination`",
        init_type=":class:`.DirectionalIllumination` or "
        ":class:`.ConstantIllumination` or dict",
        default=":class:`DirectionalIllumination() <.DirectionalIllumination>`",
    )

    kdict: KernelDict = documented(
        attrs.field(factory=KernelDict, converter=KernelDict),
        doc="Additional kernel dictionary template appended to the "
        "experiment-controlled template.",
        type=".KernelDict",
        init_type="mapping",
        default="{}",
    )

    kpmap: KernelSceneParameterMap = documented(
        attrs.field(factory=KernelSceneParameterMap, converter=KernelSceneParameterMap),
        doc="Additional scene parameter update map template appended to the "
        "experiment-controlled template.",
        type=".KernelSceneParameterMap",
        init_type="mapping",
        default="{}",
    )

    def kdict_base(self) -> KernelDict:
        # This is inefficient and exists at the moment only for debugging purposes
        return traverse(self.scene)[0]

    def kdict_full(self) -> KernelDict:
        # Return the user-defined kdict template merged with additional scene
        # element contributions
        kdict = self.kdict_base()
        kdict.update(self.kdict)
        return kdict

    def kpmap_base(self) -> KernelSceneParameterMap:
        # This is inefficient and exists at the moment only for debugging purposes
        return traverse(self.scene)[1]

    def kpmap_full(self) -> KernelSceneParameterMap:
        # Return the user-defined kpmap template merged with additional scene
        # element contributions
        kpmap = self.kpmap_base()
        kpmap.update(self.kpmap)
        return kpmap

    def _dataset_metadata(self, measure: Measure) -> dict[str, str]:
        """
        Generate additional metadata applied to dataset after post-processing.

        Parameters
        ----------
        measure : :class:`.Measure`
            Measure for which the metadata is created.

        Returns
        -------
        dict[str, str]
            Metadata to be attached to the produced dataset.
        """

        return {
            "convention": "CF-1.10",
            "source": f"eradiate, version {eradiate.__version__}",
            "history": f"{datetime.datetime.now(datetime.UTC).replace(microsecond=0).isoformat()}"
            f" - data creation - {self.__class__.__name__}.postprocess()",
            "references": "",
        }

    def spectral_indices(self, measure_index: int) -> t.Generator[SpectralIndex]:
        """
        Generate spectral indices for a given measure.

        Parameters
        ----------
        measure_index : int
            Measure index for which spectral indices are generated.

        Yields
        ------
        :class:`.SpectralIndex`
            Spectral index.
        """

        if eradiate.mode().is_mono:
            spectral_grid: MonoSpectralGrid = self.spectral_grids[measure_index]

            def generator():
                yield from spectral_grid.walk_indices()

        elif eradiate.mode().is_ckd:
            spectral_grid: CKDSpectralGrid = self.spectral_grids[measure_index]
            quad_config = self.ckd_quad_config
            try:
                abs_db = self.atmosphere.abs_db
            except (
                AttributeError
            ):  # There is either no atmosphere or no absorption database
                abs_db = None

            def generator():
                yield from spectral_grid.walk_indices(quad_config, abs_db)
        else:
            raise UnsupportedModeError

        yield from generator()

    def context_init(self):
        # Inherit docstring

        return KernelContext(
            si=self.spectral_indices(0).__next__(), kwargs=self._context_kwargs()
        )

    @abstractmethod
    def _context_kwargs(self) -> dict[str, t.Any]:
        pass

    def contexts(self, measures: None | int | list[int] = None) -> list[KernelContext]:
        # Inherit docstring

        # Collect contexts from all measures
        sis = []

        if measures is None:
            measures = list(range(len(self.measures)))

        if isinstance(measures, int):
            measures = [measures]

        for measure_index in measures:
            _si = list(self.spectral_indices(measure_index))
            sis.extend(_si)

        # Sort and remove duplicates
        key = {
            MonoSpectralIndex: lambda si: si.w.m,
            CKDSpectralIndex: lambda si: (si.w.m, si.g),
        }[SpectralIndex.subtypes.resolve()]

        sis = deduplicate_sorted(
            sorted(sis, key=key), cmp=lambda x, y: key(x) == key(y)
        )

        kwargs = self._context_kwargs()
        return [KernelContext(si, kwargs=kwargs) for si in sis]

    @property
    @abstractmethod
    def scene_objects(self) -> dict[str, SceneElement]:
        """
        Return a dictionary of string identifiers to the :class:`.SceneElement`
        instances generated by this experiment.

        Notes
        -----
        These elements make up the core scene contents for this experiment, and
        are merged together with the contents of the ``extra_objects``
        dictionary to initialize the underlying kernel scene. If you are writing
        your own :class:`.EarthObservationExperiment` subclass, this property is
        critical.
        """
        pass

    @property
    def scene(self) -> Scene:
        """
        Return a scene object used for kernel dictionary template and parameter
        table generation.
        """
        return Scene(objects={**self.scene_objects, **self.extra_objects})

    def init(self, drop_parameters: bool = True):
        # Inherit docstring

        logger.info("Initializing kernel scene")

        kdict_template, umap_template = traverse(self.scene)
        kdict_template.update(self.kdict)
        umap_template.update(self.kpmap)
        try:
            ctx = self.context_init()
            self.mi_scene = mi_traverse(
                mi.load_dict(kdict_template.render(ctx=ctx)),
                umap_template=umap_template,
            )
        except RuntimeError as e:
            raise RuntimeError(f"(while loading kernel scene dictionary){e}") from e

        # Remove unused elements from Mitsuba scene parameter table
        if drop_parameters:
            self.mi_scene.drop_parameters()

    def process(
        self,
        measures: None | int | str | list[int | str] = None,
        spp: int = 0,
        seed_state: SeedState | None = None,
    ) -> None:
        # Inherit docstring

        # Set up Mitsuba scene
        if self.mi_scene is None:
            self.init()

        # Normalize list of processed measures
        if measures is None:
            measures = self.measures
        else:
            if isinstance(measures, (int, str)):
                measures = [measures]
            measures = [self.measures.resolve(i) for i in measures]

        # Generate kernel contexts
        measure_idxs = [self.measures.get_index(measure.id) for measure in measures]
        ctxs = self.contexts(measure_idxs)

        # Collect active sensor IDs
        active_sensors = [measure.sensor_id for measure in measures]
        mi_sensors = self.mi_scene.obj.sensors()
        active_sensors = [
            i for i, sensor in enumerate(mi_sensors) if sensor.id() in active_sensors
        ]

        # Run Mitsuba for each context
        logger.info("Launching simulation")

        mi_results = mi_render(
            self.mi_scene,
            ctxs=ctxs,
            sensors=active_sensors,
            seed_state=seed_state,
            spp=spp,
        )

        # Assign collected results to the appropriate measure
        sensor_to_measure: dict[str, Measure] = {
            measure.sensor_id: measure for measure in measures
        }

        def convert_to_y_format(img):
            img_np = np.array(img, copy=False)[:, :, [0]]
            return mi.Bitmap(img_np, mi.Bitmap.PixelFormat.Y)

        # Map bitmap names to result names
        mapping = {}
        if self.integrator.stokes:
            stokes = ["S0", "S1", "S2", "S3"]
            iquv = ["I", "Q", "U", "V"]

            if self.integrator.moment:
                stokes = ["nested." + s for s in stokes]
                stokes += ["m2_" + s for s in stokes]
                iquv += ["m2_" + s for s in iquv]

            for s, i in zip(stokes, iquv):
                mapping[s] = i

        else:
            mapping = {"<root>": "bitmap"}
            if self.integrator.moment:
                mapping["m2_nested"] = "m2"

        # gather results and info from measures
        for ctx_index, spectral_group_dict in mi_results.items():
            for sensor_id, mi_bitmap in spectral_group_dict.items():
                measure = sensor_to_measure[sensor_id]
                result_imgs = {"spp": spp if spp > 0 else measure.spp}

                splits = mi_bitmap.split()
                for split in splits:
                    if split[0] in mapping:
                        img = split[1]
                        # convert any result that has more than one channel
                        if img.pixel_format() != mi.Bitmap.PixelFormat.Y:
                            img = convert_to_y_format(img)
                        result_imgs[mapping[split[0]]] = img

                measure.mi_results[ctx_index] = result_imgs

    def postprocess(self, measures: None | int | list[int] = None) -> None:
        # Inherit docstring
        logger.info("Post-processing results")

        if measures is None:
            measures = list(range(len(self.measures)))
        else:
            if isinstance(measures, (int, str)):
                measures = [self.measures.get_index(measures)]

        # Run pipelines
        for i in measures:
            measure = self.measures[i]
            drv: Driver = self.pipeline(measure)
            inputs = self._pipeline_inputs(i)
            outputs = pl.outputs(drv)
            result = drv.execute(final_vars=outputs, inputs=inputs)
            self.results[measure.id] = xr.Dataset({var: result[var] for var in outputs})

    def pipeline(self, measure: Measure | int | str) -> Driver:
        # Inherit docstring
        if isinstance(measure, (int, str)):
            measure = self.measures.resolve(measure)
        config = pl.config(measure, integrator=self.integrator)
        return eradiate.pipelines.driver(config, "eradiate.pipelines.definitions.core")

    def _pipeline_inputs(self, i_measure: int):
        # This convenience function collects pipeline inputs for a specific measure

        measure = self.measures[i_measure]
        result = {
            "bitmaps": measure.mi_results,
            "spectral_grid": self.spectral_grids[i_measure],
            "ckd_quads": self.ckd_quads[i_measure],
            "illumination": self.illumination,
            "srf": measure.srf,
        }

        config = pl.config(measure)
        if config.get("add_viewing_angles", False):
            result["angles"] = measure.viewing_angles.m_as(ureg.deg)
        else:
            result["viewing_angles"] = None

        return result


# ------------------------------------------------------------------------------
#                              Experiment runner
# ------------------------------------------------------------------------------


def run(
    exp: Experiment,
    measures: None | int | str | list[int | str] = None,
    spp: int = 0,
    seed_state: SeedState | None = None,
) -> xr.Dataset | dict[str, xr.Dataset]:
    """
    Run an Eradiate experiment. This function performs kernel scene assembly,
    runs the computation and post-processes the raw results. The output consists
    of one or several xarray datasets.

    Parameters
    ----------
    exp : Experiment
        Reference to the experiment object which will be processed.

    measures : int or str or list of int or str, optional
        Indices of the measures that will be processed. By default, all measures
        are processed.

    spp : int, optional, default: 0
        Optional parameter to override the number of samples per pixel for all
        computed measures. If set to 0, the configured value for each measure
        takes precedence.

    seed_state : :class:`.SeedState`, optional
            Seed state used to generate seeds to initialize Mitsuba's RNG at
            every iteration of the parametric loop. If unset, Eradiate's
            :attr:`root seed state <.root_seed_state>` is used.

    Returns
    -------
    Dataset or dict[str, Dataset]
        If a single measure is processed, a single xarray dataset is returned.
        If several measures are processed, a dictionary mapping measure IDs to
        the corresponding result dataset is returned.

    Notes
    -----
    * Successive calls to this function with different measures will not reset
      the :attr:`Experiment.results` dictionary.
    * Successive calls with already processed measures will overwrite prior
      results.
    """
    if measures is None:
        measures = list(range(len(exp.measures)))
    if isinstance(measures, (int, str)):
        measures = [measures]

    exp.process(spp=spp, measures=measures, seed_state=seed_state)
    exp.postprocess(measures=measures)

    measure_ids = [exp.measures.get_id(m) for m in measures]
    return (
        {x: exp.results[x] for x in measure_ids}
        if len(measure_ids) > 1
        else exp.results[measure_ids[0]]
    )



from __future__ import annotations

from abc import ABC

import attrs

from ..core import NodeSceneElement, SceneElement
from ..._factory import Factory
from ...attrs import define, documented, get_doc

phase_function_factory = Factory()
phase_function_factory.register_lazy_batch(
    [
        (
            "_blend.BlendPhaseFunction",
            "blend_phase",
            {},
        ),
        (
            "_hg.HenyeyGreensteinPhaseFunction",
            "hg",
            {},
        ),
        (
            "_isotropic.IsotropicPhaseFunction",
            "isotropic",
            {},
        ),
        (
            "_rayleigh.RayleighPhaseFunction",
            "rayleigh",
            {},
        ),
        (
            "_tabulated.TabulatedPhaseFunction",
            "tab_phase",
            {},
        ),
    ],
    cls_prefix="eradiate.scenes.phase",
)


@define(eq=False, slots=False)
class PhaseFunction(NodeSceneElement, ABC):
    """
    An abstract base class defining common facilities for all phase functions.
    """

    id: str | None = documented(
        attrs.field(
            default="phase",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(SceneElement, "id", "doc"),
        type=get_doc(SceneElement, "id", "type"),
        init_type=get_doc(SceneElement, "id", "init_type"),
        default='"phase"',
    )



from __future__ import annotations

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr

from ._core import Spectrum
from ... import converters, validators
from ...attrs import define, documented
from ...kernel import DictParameter, KernelSceneParameterFlags, SceneParameter
from ...units import PhysicalQuantity, to_quantity
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg

_trapezoid = np.trapezoid if int(np.__version__.split(".")[0]) >= 2 else np.trapz


@define(eq=False, slots=False, init=False)
class InterpolatedSpectrum(Spectrum):
    """
    Linearly interpolated spectrum [``interpolated``].

    .. admonition:: Class method constructors

       .. autosummary::

          from_dataarray

    Warnings
    --------
    In all modes, queries outside the spectral domain covered by the input data
    return 0.

    Notes
    -----
    * Interpolation uses :func:`numpy.interp`. In both ``mono_*`` and ``ckd_*``
      modes, the spectrum is evaluated at the spectral context wavelength.
    * Integration performed with the :meth:`.integral` method uses a trapezoid
      rule.
    """

    wavelengths: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("wavelength"),
            converter=[
                np.atleast_1d,
                pinttr.converters.to_units(ucc.deferred("wavelength")),
            ],
            kw_only=True,
        ),
        doc="Wavelengths defining the interpolation grid. Values must be "
        "monotonically increasing.",
        type="quantity",
    )

    @wavelengths.validator
    def _wavelengths_validator(self, attribute, value):
        # Wavelengths must be monotonically increasing
        if not np.all(np.diff(value) > 0):
            raise ValueError("wavelengths must be monotonically increasing")

        # Check values
        if np.any(np.isnan(value)):
            raise ValueError("Detected NaN in 'wavelengths'")

    values: pint.Quantity = documented(
        attrs.field(
            converter=converters.on_quantity(np.atleast_1d),
            kw_only=True,
        ),
        doc="Spectrum values. If a quantity is passed, units are "
        "checked for consistency. If a unitless array is passed, it is "
        "automatically converted to appropriate default configuration units, "
        "depending on the value of the ``quantity`` field. If no quantity is "
        "specified, this field can be a unitless value.",
        type="quantity",
        init_type="array-like",
    )

    @values.validator
    def _values_validator(self, attribute, value):
        if self.quantity is not None:
            if not isinstance(self.values, pint.Quantity):
                raise ValueError(
                    f"while validating '{attribute.name}': expected a Pint "
                    "quantity compatible with quantity field value "
                    f"'{self.quantity}', got a unitless value instead"
                )

            expected_units = ucc.get(self.quantity)
            if not pinttr.util.units_compatible(expected_units, value.units):
                raise pinttr.exceptions.UnitsError(
                    value.units,
                    expected_units,
                    extra_msg=f"while validating '{attribute.name}', got units "
                    f"'{value.units}' incompatible with quantity {self.quantity} "
                    f"(expected '{expected_units}')",
                )

    @values.validator
    @wavelengths.validator
    def _values_wavelengths_validator(self, attribute, value):
        # Check that attribute is an array
        validators.on_quantity(attrs.validators.instance_of(np.ndarray))(
            self, attribute, value
        )

        # Check size
        if value.ndim > 1:
            raise ValueError(
                f"while validating '{attribute.name}': '{attribute.name}' must "
                "be a 1D array"
            )

        if len(value) < 2:
            raise ValueError(
                f"while validating '{attribute.name}': '{attribute.name}' must "
                f"have length >= 2"
            )

        if self.wavelengths.shape != self.values.shape:
            raise ValueError(
                f"while validating '{attribute.name}': 'wavelengths' and 'values' "
                f"must have the same shape, got {self.wavelengths.shape} and "
                f"{self.values.shape}"
            )

        # Check values
        if np.any(np.isnan(value)):
            raise ValueError("Detected NaN in 'values'")

    def __init__(
        self,
        id: str | None = None,
        quantity: PhysicalQuantity | str | None = None,
        *,
        wavelengths: np.typing.ArrayLike,
        values: np.typing.ArrayLike,
    ):
        # If a quantity is set and a unitless value is passed, it is
        # automatically applied appropriate units
        if quantity is not None and not isinstance(values, pint.Quantity):
            values = pinttr.util.ensure_units(values, ucc.get(quantity))

        self.__attrs_init__(
            id=id, quantity=quantity, wavelengths=wavelengths, values=values
        )

    def update(self):
        # Sort by increasing wavelength (required by numpy.interp in eval_mono)
        idx = np.argsort(self.wavelengths)
        self.wavelengths = self.wavelengths[idx]
        self.values = self.values[idx]

    @classmethod
    def from_dataarray(
        cls,
        id: str | None = None,
        quantity: str | PhysicalQuantity | None = None,
        *,
        dataarray: xr.DataArray,
    ) -> InterpolatedSpectrum:
        """
        Construct an interpolated spectrum from an xarray data array.

        Parameters
        ----------
        id : str, optional
            Optional object identifier.

        quantity : str or .PhysicalQuantity, optional
            If set, quantity represented by the spectrum. This parameter and
            spectrum units must be consistent. This parameter takes precedence
            over the ``quantity`` field of the data array.

        dataarray : DataArray
            An :class:`xarray.DataArray` instance complying to the spectrum data
            array format (see *Notes*).

        Notes
        -----

        * Expected data format:

          **Coordinates (\\* means also dimension)**

          * ``*w`` (float): wavelength (units specified as a ``units`` attribute).

          **Metadata**

          * ``quantity`` (str): physical quantity which the data describes (see
            :meth:`.PhysicalQuantity.spectrum` for allowed values), optional.
          * ``units`` (str): units of spectrum values (must be consistent with
            ``quantity``).
        """
        kwargs = {}

        if id is not None:
            kwargs[id] = id

        if quantity is None:
            try:
                kwargs["quantity"] = dataarray.attrs["quantity"]
            except KeyError:
                pass
        else:
            kwargs["quantity"] = quantity

        values = to_quantity(dataarray)
        wavelengths = to_quantity(dataarray.w)

        return cls(**kwargs, values=values, wavelengths=wavelengths)

    def eval_mono(self, w: pint.Quantity) -> pint.Quantity:
        return np.interp(w, self.wavelengths, self.values, left=0.0, right=0.0)

    def eval_ckd(self, w: pint.Quantity, g: float) -> pint.Quantity:
        return self.eval_mono(w=w)

    def integral(self, wmin: pint.Quantity, wmax: pint.Quantity) -> pint.Quantity:
        # Inherit docstring

        # Collect spectral coordinates
        wavelength_units = self.wavelengths.units
        s_w = self.wavelengths.m
        s_wmin = s_w.min()
        s_wmax = s_w.max()

        # Select all spectral mesh vertices between wmin and wmax, as well as
        # the bounds themselves
        wmin = wmin.m_as(wavelength_units)
        wmax = wmax.m_as(wavelength_units)
        w = [wmin, *s_w[np.where(np.logical_and(wmin < s_w, s_w < wmax))[0]], wmax]

        # Make explicit the fact that the function underlying this spectrum has
        # a finite support by padding the s_wmin and s_wmax values with a very
        # small margin
        eps = 1e-12  # nm

        try:
            w.insert(w.index(s_wmin), s_wmin - eps)
        except ValueError:
            pass

        try:
            w.insert(w.index(s_wmax) + 1, s_wmax + eps)
        except ValueError:
            pass

        # Evaluate spectrum at wavelengths
        interp = self.eval_mono(w * wavelength_units)

        # Compute integral
        integral = _trapezoid(interp, w)

        # Apply units
        return integral * ureg.dimensionless * wavelength_units

    @property
    def template(self) -> dict:
        # Inherit docstring

        return {
            "type": "uniform",
            "value": DictParameter(
                func=lambda ctx: float(self.eval(ctx.si).m_as(uck.get(self.quantity)))
            ),
        }

    @property
    def params(self) -> dict:
        # Inherit docstring

        return {
            "value": SceneParameter(
                func=lambda ctx: float(self.eval(ctx.si).m_as(uck.get(self.quantity))),
                flags=KernelSceneParameterFlags.SPECTRAL,
            )
        }



from __future__ import annotations

import attrs

import eradiate

from ._core import Integrator
from ...attrs import define, documented


@define(eq=False, slots=False)
class MonteCarloIntegrator(Integrator):
    """
    Base class for integrator elements wrapping kernel classes
    deriving from
    :class:`mitsuba.MonteCarloIntegrator`.

    .. warning:: This class should not be instantiated.
    """

    max_depth: int | None = documented(
        attrs.field(default=None, converter=attrs.converters.optional(int)),
        doc="Longest path depth in the generated measure data (where -1 "
        "corresponds to ∞). A value of 1 will display only visible emitters. 2 "
        "computes only direct illumination (no multiple scattering), etc. If "
        "unset, the kernel default value (-1) is used.",
        type="int or None",
        init_type="int, optional",
    )

    rr_depth: int | None = documented(
        attrs.field(default=None, converter=attrs.converters.optional(int)),
        doc="Minimum path depth after which the implementation starts applying "
        "the Russian roulette path termination criterion. If unset, the kernel "
        "default value (5) is used.",
        type="int or None",
        init_type="int, optional",
    )

    hide_emitters: bool | None = documented(
        attrs.field(default=None, converter=attrs.converters.optional(bool)),
        doc="Hide directly visible emitters. If unset, the kernel default "
        "value (``false``) is used.",
        type="bool or None",
        init_type="bool, optional",
    )

    @property
    def kernel_type(self) -> str:
        raise NotImplementedError

    @property
    def template(self) -> dict:
        result = {"type": self.kernel_type}

        if self.timeout is not None:
            result["timeout"] = self.timeout
        if self.max_depth is not None:
            result["max_depth"] = self.max_depth
        if self.rr_depth is not None:
            result["rr_depth"] = self.rr_depth
        if self.hide_emitters is not None:
            result["hide_emitters"] = self.hide_emitters

        if self.stokes and not eradiate.mode().is_polarized:
            raise RuntimeError("stokes should only be set to True in polarized mode.")

        if self.stokes:
            result = {
                "type": "stokes",
                "integrator": result,
                "meridian_align": self.meridian_align,
            }

        if self.moment:
            result = {"type": "moment", "nested": result}

        return result


@define(eq=False, slots=False)
class PathIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the path tracer kernel plugin [``path``].

    This integrator samples paths using random walks starting from the sensor.
    It supports multiple scattering and does not account for volume
    interactions.
    """

    @property
    def kernel_type(self) -> str:
        return "path"


@define(eq=False, slots=False)
class VolPathIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the volumetric path tracer kernel plugin [``volpath``].

    This integrator samples paths using random walks starting from the sensor.
    It supports multiple scattering and accounts for volume interactions.
    """

    @property
    def kernel_type(self) -> str:
        return "volpath"


@define(eq=False, slots=False)
class VolPathMISIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the volumetric path tracer kernel plugin [``volpathmis``].

    This plugin implements spectral multiple importance sampling
    :cite:`Miller2019NullscatteringPathIntegral`.
    """

    use_spectral_mis = attrs.field(
        default=None, converter=attrs.converters.optional(bool)
    )

    @property
    def kernel_type(self) -> str:
        return "volpathmis"

    @property
    def template(self) -> dict:
        result = super().template
        if self.use_spectral_mis is not None:
            result["use_spectral_mis"] = self.use_spectral_mis

        return result


@define(eq=False, slots=False)
class PiecewiseVolPathIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the piecewise volumetric path tracer kernel plugin [``piecewise_volpath``].

    This integrator samples paths using random walks starting from the sensor.
    It supports multiple scattering and accounts for 1D volume interactions.
    """

    @property
    def kernel_type(self) -> str:
        return "piecewise_volpath"



from __future__ import annotations

import glob
import importlib.resources
from pathlib import Path

import cerberus
import xarray as xr
from ruamel.yaml import YAML

from ..units import unit_registry as ureg
from ..units import units_compatible


def _load_rules():
    cerberus.rules_set_registry.extend([("integer", {"type": "integer"})])
    cerberus.rules_set_registry.extend([("string", {"type": "string"})])


def _load_schemas() -> None:
    schema_paths = sorted(
        Path(x)
        for x in glob.glob(
            str(importlib.resources.files("eradiate.data.schemas").joinpath("*.yml"))
        )
    )  # TODO: check if that would work on Windows

    yaml = YAML(typ="safe")
    for path in schema_paths:
        name = path.stem
        with open(path, "r") as f:
            schema = yaml.load(f)
        cerberus.schema_registry.add(name, schema)


_load_rules()
_load_schemas()


class DatasetValidator(cerberus.Validator):
    """
    This class implements xarray dataset validation. In addition to providing
    a simple interface to validate the structure of a dataset against a Cerberus
    schema, it provides validation rules
    """

    def _validate_equal_list(self, constraint, field, value):
        """
        Check if the value is equal to a specific list.

        The rule's arguments are validated against this schema:
        {"type": "list"}
        """
        if list(value) != list(constraint):
            self._error(field, f"Must be equal to {constraint}")

    def _validate_units_compatible(self, constraint, field, value):
        """
        Check if a 'units' attribute field has a value that is compatible with
        specific units. The underlying implementation uses
        :func:`.units_compatible` and will consequently not return a false
        positive if validating angle units against dimensionless quantities.

        The rule's arguments are validated against this schema:
        {"type": "string"}
        """
        try:
            u_value = ureg.Unit(value)
        except Exception:
            self._error(field, f"Cannot convert {repr(value)} to valid units")
            return

        u_constraint = ureg.Unit(constraint)
        if not units_compatible(u_constraint, u_value):
            self._error(field, f"Must be units compatible with {constraint}")

    def validate(self, document, schema=None, update=False, normalize=True):
        # Inherit docstring
        if isinstance(document, xr.Dataset):
            document = document.to_dict(data=False)
        super().validate(document, schema=schema, update=update, normalize=normalize)


def list_schemas() -> list[str]:
    """
    List currently registered Cerberus schemas.
    """
    return sorted(cerberus.schema_registry._storage.keys())
